"""
# Copyright (c) 2017 The Khronos Group Inc.
# Copyright (c) 2016 Oculus VR, LLC.
#
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License")
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http:#www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: J.M.P. van Waveren
#
"""

from ctypes import addressof, c_float, Structure
import enum
import math
import numpy
from typing import Sequence

from .typedefs import *


class GraphicsAPI(enum.Enum):
    VULKAN = 0,
    OPENGL = 1,
    OPENGL_ES = 2,
    D3D = 3,


# Column-major, pre-multiplied. This type does not exist in the OpenXR API and is provided for convenience.
class Matrix4x4f(Structure):
    _fields_ = [("m", c_float * 16), ]

    def __init__(self):
        super().__init__()
        self._numpy = None

    def __matmul__(self, other) -> "Matrix4x4f":
        return self.multiply(other)

    def as_numpy(self):
        if not hasattr(self, "_numpy") or self._numpy is None:
            # Just in time construction
            buffer = (c_float * 16).from_address(addressof(self))
            buffer._wrapper = self  # To link lifetime of buffer to self
            self._numpy = numpy.ctypeslib.as_array(buffer)
        return self._numpy

    @staticmethod
    def create_from_quaternion(quat: Quaternionf) -> "Matrix4x4f":
        """ Creates a matrix from a quaternion. """
        x2 = quat.x + quat.x
        y2 = quat.y + quat.y
        z2 = quat.z + quat.z

        xx2 = quat.x * x2
        yy2 = quat.y * y2
        zz2 = quat.z * z2

        yz2 = quat.y * z2
        wx2 = quat.w * x2
        xy2 = quat.x * y2
        wz2 = quat.w * z2
        xz2 = quat.x * z2
        wy2 = quat.w * y2

        result = Matrix4x4f()
        result.m[0] = 1.0 - yy2 - zz2
        result.m[1] = xy2 + wz2
        result.m[2] = xz2 - wy2
        result.m[3] = 0.0

        result.m[4] = xy2 - wz2
        result.m[5] = 1.0 - xx2 - zz2
        result.m[6] = yz2 + wx2
        result.m[7] = 0.0

        result.m[8] = xz2 + wy2
        result.m[9] = yz2 - wx2
        result.m[10] = 1.0 - xx2 - yy2
        result.m[11] = 0.0

        result.m[12] = 0.0
        result.m[13] = 0.0
        result.m[14] = 0.0
        result.m[15] = 1.0
        return result

    @staticmethod
    def create_projection(graphics_api: GraphicsAPI, tan_angle_left: float, tan_angle_right: float, tan_angle_up: float,
                          tan_angle_down: float, near_z: float, far_z: float) -> "Matrix4x4f":
        """
        # Creates a projection matrix based on the specified dimensions.
        # The projection matrix transforms -Z=forward, +Y=up, +X=right to the appropriate clip space for the graphics API.
        # The far plane is placed at infinity if far_z <= near_z.
        # An infinite projection matrix is preferred for rasterization because, except for
        # things *right* up against the near plane, it always provides better precision:
        #              "Tightening the Precision of Perspective Rendering"
        #              Paul Upchurch, Mathieu Desbrun
        #              Journal of Graphics Tools, Volume 16, Issue 1, 2012
        """
        tan_angle_width = tan_angle_right - tan_angle_left
        # Set to tan_angle_down - tan_angle_up for a clip space with positive Y down (Vulkan).
        # Set to tan_angle_up - tan_angle_down for a clip space with positive Y up (OpenGL / D3D / Metal).
        tan_angle_height = (tan_angle_down - tan_angle_up) if graphics_api == GraphicsAPI.VULKAN else (tan_angle_up - tan_angle_down)
        # Set to near_z for a [-1,1] Z clip space (OpenGL / OpenGL ES).
        # Set to zero for a [0,1] Z clip space (Vulkan / D3D / Metal).
        offset_z = near_z if graphics_api == GraphicsAPI.OPENGL or graphics_api == GraphicsAPI.OPENGL_ES else 0
        result = Matrix4x4f()
        if far_z <= near_z:
            # place the far plane at infinity
            result.m[0] = 2.0 / tan_angle_width
            result.m[4] = 0.0
            result.m[8] = (tan_angle_right + tan_angle_left) / tan_angle_width
            result.m[12] = 0.0

            result.m[1] = 0.0
            result.m[5] = 2.0 / tan_angle_height
            result.m[9] = (tan_angle_up + tan_angle_down) / tan_angle_height
            result.m[13] = 0.0

            result.m[2] = 0.0
            result.m[6] = 0.0
            result.m[10] = -1.0
            result.m[14] = -(near_z + offset_z)

            result.m[3] = 0.0
            result.m[7] = 0.0
            result.m[11] = -1.0
            result.m[15] = 0.0
        else:
            # normal projection
            result.m[0] = 2.0 / tan_angle_width
            result.m[4] = 0.0
            result.m[8] = (tan_angle_right + tan_angle_left) / tan_angle_width
            result.m[12] = 0.0

            result.m[1] = 0.0
            result.m[5] = 2.0 / tan_angle_height
            result.m[9] = (tan_angle_up + tan_angle_down) / tan_angle_height
            result.m[13] = 0.0

            result.m[2] = 0.0
            result.m[6] = 0.0
            result.m[10] = -(far_z + offset_z) / (far_z - near_z)
            result.m[14] = -(far_z * (near_z + offset_z)) / (far_z - near_z)

            result.m[3] = 0.0
            result.m[7] = 0.0
            result.m[11] = -1.0
            result.m[15] = 0.0
        return result

    @staticmethod
    def create_projection_fov(graphics_api: GraphicsAPI, fov: Fovf, near_z: float, far_z: float) -> "Matrix4x4f":
        """ Creates a projection matrix based on the specified FOV. """
        tan_left = math.tan(fov.angle_left)
        tan_right = math.tan(fov.angle_right)
        tan_down = math.tan(fov.angle_down)
        tan_up = math.tan(fov.angle_up)
        return Matrix4x4f.create_projection(graphics_api, tan_left, tan_right, tan_up, tan_down, near_z, far_z)

    @staticmethod
    def create_scale(x: float, y: float, z: float) -> "Matrix4x4f":
        """ Creates a scale matrix. """
        result = Matrix4x4f()
        result.m[0] = x
        result.m[1] = 0.0
        result.m[2] = 0.0
        result.m[3] = 0.0
        result.m[4] = 0.0
        result.m[5] = y
        result.m[6] = 0.0
        result.m[7] = 0.0
        result.m[8] = 0.0
        result.m[9] = 0.0
        result.m[10] = z
        result.m[11] = 0.0
        result.m[12] = 0.0
        result.m[13] = 0.0
        result.m[14] = 0.0
        result.m[15] = 1.0
        return result

    @staticmethod
    def create_translation(x: float, y: float, z: float) -> "Matrix4x4f":
        """ Creates a translation matrix. """
        result = Matrix4x4f()
        result.m[0] = 1.0
        result.m[1] = 0.0
        result.m[2] = 0.0
        result.m[3] = 0.0
        result.m[4] = 0.0
        result.m[5] = 1.0
        result.m[6] = 0.0
        result.m[7] = 0.0
        result.m[8] = 0.0
        result.m[9] = 0.0
        result.m[10] = 1.0
        result.m[11] = 0.0
        result.m[12] = x
        result.m[13] = y
        result.m[14] = z
        result.m[15] = 1.0
        return result

    @staticmethod
    def create_translation_rotation_scale(translation: Vector3f, rotation: Quaternionf, scale: Sequence[float]) -> "Matrix4x4f":
        """ Creates a combined translation(rotation(scale(object))) matrix. """
        scale_matrix = Matrix4x4f.create_scale(*scale)
        rotation_matrix = Matrix4x4f.create_from_quaternion(rotation)
        translation_matrix = Matrix4x4f.create_translation(*translation)
        combined_matrix = rotation_matrix @ scale_matrix
        return translation_matrix @ combined_matrix

    def multiply(self, b: "Matrix4x4f") -> "Matrix4x4f":
        """ Use left-multiplication to accumulate transformations. """
        result = Matrix4x4f()
        result.m[0] = self.m[0] * b.m[0] + self.m[4] * b.m[1] + self.m[8] * b.m[2] + self.m[12] * b.m[3]
        result.m[1] = self.m[1] * b.m[0] + self.m[5] * b.m[1] + self.m[9] * b.m[2] + self.m[13] * b.m[3]
        result.m[2] = self.m[2] * b.m[0] + self.m[6] * b.m[1] + self.m[10] * b.m[2] + self.m[14] * b.m[3]
        result.m[3] = self.m[3] * b.m[0] + self.m[7] * b.m[1] + self.m[11] * b.m[2] + self.m[15] * b.m[3]

        result.m[4] = self.m[0] * b.m[4] + self.m[4] * b.m[5] + self.m[8] * b.m[6] + self.m[12] * b.m[7]
        result.m[5] = self.m[1] * b.m[4] + self.m[5] * b.m[5] + self.m[9] * b.m[6] + self.m[13] * b.m[7]
        result.m[6] = self.m[2] * b.m[4] + self.m[6] * b.m[5] + self.m[10] * b.m[6] + self.m[14] * b.m[7]
        result.m[7] = self.m[3] * b.m[4] + self.m[7] * b.m[5] + self.m[11] * b.m[6] + self.m[15] * b.m[7]

        result.m[8] = self.m[0] * b.m[8] + self.m[4] * b.m[9] + self.m[8] * b.m[10] + self.m[12] * b.m[11]
        result.m[9] = self.m[1] * b.m[8] + self.m[5] * b.m[9] + self.m[9] * b.m[10] + self.m[13] * b.m[11]
        result.m[10] = self.m[2] * b.m[8] + self.m[6] * b.m[9] + self.m[10] * b.m[10] + self.m[14] * b.m[11]
        result.m[11] = self.m[3] * b.m[8] + self.m[7] * b.m[9] + self.m[11] * b.m[10] + self.m[15] * b.m[11]

        result.m[12] = self.m[0] * b.m[12] + self.m[4] * b.m[13] + self.m[8] * b.m[14] + self.m[12] * b.m[15]
        result.m[13] = self.m[1] * b.m[12] + self.m[5] * b.m[13] + self.m[9] * b.m[14] + self.m[13] * b.m[15]
        result.m[14] = self.m[2] * b.m[12] + self.m[6] * b.m[13] + self.m[10] * b.m[14] + self.m[14] * b.m[15]
        result.m[15] = self.m[3] * b.m[12] + self.m[7] * b.m[13] + self.m[11] * b.m[14] + self.m[15] * b.m[15]
        return result

    """ Calculates the inverse of a rigid body transform. """
    def invert_rigid_body(self) -> "Matrix4x4f":
        result = Matrix4x4f()
        result.m[0] = self.m[0]
        result.m[1] = self.m[4]
        result.m[2] = self.m[8]
        result.m[3] = 0.0
        result.m[4] = self.m[1]
        result.m[5] = self.m[5]
        result.m[6] = self.m[9]
        result.m[7] = 0.0
        result.m[8] = self.m[2]
        result.m[9] = self.m[6]
        result.m[10] = self.m[10]
        result.m[11] = 0.0
        result.m[12] = -(self.m[0] * self.m[12] + self.m[1] * self.m[13] + self.m[2] * self.m[14])
        result.m[13] = -(self.m[4] * self.m[12] + self.m[5] * self.m[13] + self.m[6] * self.m[14])
        result.m[14] = -(self.m[8] * self.m[12] + self.m[9] * self.m[13] + self.m[10] * self.m[14])
        result.m[15] = 1.0
        return result


__all__ = [
    "GraphicsAPI",
    "Matrix4x4f",
]



from ctypes import Array, byref, c_char, c_int, c_uint32, c_void_p, Structure
import ctypes

from . import raw_functions  # Side effect of defining low-level c signatures
from .enums import *
from .exception import check_result
from .typedefs import *


class ExtensionProperties(Structure):
    _fields_ = [
        ("type", c_int),  #  StructureType),
        ("next", c_void_p),
        ("extension_name", (c_char * 128)),
        ("extension_version", c_uint32),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(
            StructureType.EXTENSION_PROPERTIES.value,
            None,
            *args, **kwargs,
        )

    def __bytes__(self):
        return self.extension_name

    def __eq__(self, other):
        try:
            if other.type != self.type:
                return False
            if other.extension_version != self.extension_version:
                return False
        except AttributeError as exc:
            pass  # That's OK, objects without those attributes can use string comparison
        return str(other) == str(self)

    def __str__(self):
        return self.extension_name.decode()

    @classmethod
    def make_array(cls, element_count: int):
        result = (cls * element_count)()
        for element in result:
            element.type = StructureType.EXTENSION_PROPERTIES.value
        return result


if False:
    raw_functions.xrEnumerateInstanceExtensionProperties.argtypes = [
        ctypes.c_char_p,  # layer_name
        c_uint32,  # property_capacity_input
        ctypes.POINTER(c_uint32),  # property_count_output
        ctypes.POINTER(ExtensionProperties),  # properties
    ]


# function transformations:
#  * brief docstring
#  * parameter docstring
#  * exception docstring
#  * output_array_parameters
#  * check result
#  * string input parameter


def enumerate_instance_extension_properties(
    layer_name: str = None,
) -> Array[ExtensionProperties]:
    """
    Returns properties of available instance extensions

    :param layer_name: is either NULL or a pointer to a string naming the API layer to retrieve extensions from,
        as returned by xrEnumerateApiLayerProperties.
    :return: an array of XrExtensionProperties

    On failure, this command raises
        XR_ERROR_VALIDATION_FAILURE
        XR_ERROR_RUNTIME_FAILURE
        XR_ERROR_OUT_OF_MEMORY
        XR_ERROR_SIZE_INSUFFICIENT
        XR_ERROR_RUNTIME_UNAVAILABLE
        XR_ERROR_API_LAYER_NOT_PRESENT
    """
    # First call returns the item count
    if layer_name is not None:
        layer_name = layer_name.encode()

    extension_count = c_uint32(0)
    fn = raw_functions.xrEnumerateInstanceExtensionProperties
    result = check_result(fn(layer_name, 0, byref(extension_count), None))
    if result.is_exception():
        raise result

    properties = ExtensionProperties.make_array(extension_count.value)

    # TODO: automatically initialize?
    # for p in properties:
    #     p.type = StructureType.EXTENSION_PROPERTIES.value

    result = check_result(
        fn(
            layer_name,
            extension_count,
            byref(extension_count),
            properties,  # Don't use byref for arrays...
        )
    )
    if result.is_exception():
        raise result

    return properties


__all__ = [
    # "enumerate_instance_extension_properties",
    # "ExtensionProperties",
]



# Warning: this file is auto-generated. Do not edit.

from ctypes import c_int, c_uint64
import enum


def succeeded(result) -> bool:
    return result >= 0


def failed(result) -> bool:
    return result < 0


def unqualified_success(result) -> bool:
    return result == 0


class DefaultEnumMeta(enum.EnumMeta):
    """
    Metaclass to allow default values in enumerations.

    https://stackoverflow.com/questions/44867597/is-there-a-way-to-specify-a-default-value-for-python-enums
    """
    default = object()

    def __call__(cls, value=default, *args, **kwargs):
        if value is DefaultEnumMeta.default:
            # Assume the first enum is default
            return next(iter(cls))
        return super().__call__(value, *args, **kwargs)


class EnumBase(enum.IntEnum, metaclass=DefaultEnumMeta):
    @staticmethod
    def ctype():
        return c_int


class FlagBase(enum.Flag, metaclass=DefaultEnumMeta):
    @staticmethod
    def ctype():
        return c_uint64


class Result(EnumBase):
    SUCCESS = 0
    TIMEOUT_EXPIRED = 1
    SESSION_LOSS_PENDING = 3
    EVENT_UNAVAILABLE = 4
    SPACE_BOUNDS_UNAVAILABLE = 7
    SESSION_NOT_FOCUSED = 8
    FRAME_DISCARDED = 9
    ERROR_VALIDATION_FAILURE = -1
    ERROR_RUNTIME_FAILURE = -2
    ERROR_OUT_OF_MEMORY = -3
    ERROR_API_VERSION_UNSUPPORTED = -4
    ERROR_INITIALIZATION_FAILED = -6
    ERROR_FUNCTION_UNSUPPORTED = -7
    ERROR_FEATURE_UNSUPPORTED = -8
    ERROR_EXTENSION_NOT_PRESENT = -9
    ERROR_LIMIT_REACHED = -10
    ERROR_SIZE_INSUFFICIENT = -11
    ERROR_HANDLE_INVALID = -12
    ERROR_INSTANCE_LOST = -13
    ERROR_SESSION_RUNNING = -14
    ERROR_SESSION_NOT_RUNNING = -16
    ERROR_SESSION_LOST = -17
    ERROR_SYSTEM_INVALID = -18
    ERROR_PATH_INVALID = -19
    ERROR_PATH_COUNT_EXCEEDED = -20
    ERROR_PATH_FORMAT_INVALID = -21
    ERROR_PATH_UNSUPPORTED = -22
    ERROR_LAYER_INVALID = -23
    ERROR_LAYER_LIMIT_EXCEEDED = -24
    ERROR_SWAPCHAIN_RECT_INVALID = -25
    ERROR_SWAPCHAIN_FORMAT_UNSUPPORTED = -26
    ERROR_ACTION_TYPE_MISMATCH = -27
    ERROR_SESSION_NOT_READY = -28
    ERROR_SESSION_NOT_STOPPING = -29
    ERROR_TIME_INVALID = -30
    ERROR_REFERENCE_SPACE_UNSUPPORTED = -31
    ERROR_FILE_ACCESS_ERROR = -32
    ERROR_FILE_CONTENTS_INVALID = -33
    ERROR_FORM_FACTOR_UNSUPPORTED = -34
    ERROR_FORM_FACTOR_UNAVAILABLE = -35
    ERROR_API_LAYER_NOT_PRESENT = -36
    ERROR_CALL_ORDER_INVALID = -37
    ERROR_GRAPHICS_DEVICE_INVALID = -38
    ERROR_POSE_INVALID = -39
    ERROR_INDEX_OUT_OF_RANGE = -40
    ERROR_VIEW_CONFIGURATION_TYPE_UNSUPPORTED = -41
    ERROR_ENVIRONMENT_BLEND_MODE_UNSUPPORTED = -42
    ERROR_NAME_DUPLICATED = -44
    ERROR_NAME_INVALID = -45
    ERROR_ACTIONSET_NOT_ATTACHED = -46
    ERROR_ACTIONSETS_ALREADY_ATTACHED = -47
    ERROR_LOCALIZED_NAME_DUPLICATED = -48
    ERROR_LOCALIZED_NAME_INVALID = -49
    ERROR_GRAPHICS_REQUIREMENTS_CALL_MISSING = -50
    ERROR_RUNTIME_UNAVAILABLE = -51
    ERROR_ANDROID_THREAD_SETTINGS_ID_INVALID_KHR = -1000003000
    ERROR_ANDROID_THREAD_SETTINGS_FAILURE_KHR = -1000003001
    ERROR_CREATE_SPATIAL_ANCHOR_FAILED_MSFT = -1000039001
    ERROR_SECONDARY_VIEW_CONFIGURATION_TYPE_NOT_ENABLED_MSFT = -1000053000
    ERROR_CONTROLLER_MODEL_KEY_INVALID_MSFT = -1000055000
    ERROR_REPROJECTION_MODE_UNSUPPORTED_MSFT = -1000066000
    ERROR_COMPUTE_NEW_SCENE_NOT_COMPLETED_MSFT = -1000097000
    ERROR_SCENE_COMPONENT_ID_INVALID_MSFT = -1000097001
    ERROR_SCENE_COMPONENT_TYPE_MISMATCH_MSFT = -1000097002
    ERROR_SCENE_MESH_BUFFER_ID_INVALID_MSFT = -1000097003
    ERROR_SCENE_COMPUTE_FEATURE_INCOMPATIBLE_MSFT = -1000097004
    ERROR_SCENE_COMPUTE_CONSISTENCY_MISMATCH_MSFT = -1000097005
    ERROR_DISPLAY_REFRESH_RATE_UNSUPPORTED_FB = -1000101000
    ERROR_COLOR_SPACE_UNSUPPORTED_FB = -1000108000
    ERROR_SPACE_COMPONENT_NOT_SUPPORTED_FB = -1000113000
    ERROR_SPACE_COMPONENT_NOT_ENABLED_FB = -1000113001
    ERROR_SPACE_COMPONENT_STATUS_PENDING_FB = -1000113002
    ERROR_SPACE_COMPONENT_STATUS_ALREADY_SET_FB = -1000113003
    ERROR_UNEXPECTED_STATE_PASSTHROUGH_FB = -1000118000
    ERROR_FEATURE_ALREADY_CREATED_PASSTHROUGH_FB = -1000118001
    ERROR_FEATURE_REQUIRED_PASSTHROUGH_FB = -1000118002
    ERROR_NOT_PERMITTED_PASSTHROUGH_FB = -1000118003
    ERROR_INSUFFICIENT_RESOURCES_PASSTHROUGH_FB = -1000118004
    ERROR_UNKNOWN_PASSTHROUGH_FB = -1000118050
    ERROR_RENDER_MODEL_KEY_INVALID_FB = -1000119000
    RENDER_MODEL_UNAVAILABLE_FB = 1000119020
    ERROR_MARKER_NOT_TRACKED_VARJO = -1000124000
    ERROR_MARKER_ID_INVALID_VARJO = -1000124001
    ERROR_SPATIAL_ANCHOR_NAME_NOT_FOUND_MSFT = -1000142001
    ERROR_SPATIAL_ANCHOR_NAME_INVALID_MSFT = -1000142002


class StructureType(EnumBase):
    UNKNOWN = 0
    API_LAYER_PROPERTIES = 1
    EXTENSION_PROPERTIES = 2
    INSTANCE_CREATE_INFO = 3
    SYSTEM_GET_INFO = 4
    SYSTEM_PROPERTIES = 5
    VIEW_LOCATE_INFO = 6
    VIEW = 7
    SESSION_CREATE_INFO = 8
    SWAPCHAIN_CREATE_INFO = 9
    SESSION_BEGIN_INFO = 10
    VIEW_STATE = 11
    FRAME_END_INFO = 12
    HAPTIC_VIBRATION = 13
    EVENT_DATA_BUFFER = 16
    EVENT_DATA_INSTANCE_LOSS_PENDING = 17
    EVENT_DATA_SESSION_STATE_CHANGED = 18
    ACTION_STATE_BOOLEAN = 23
    ACTION_STATE_FLOAT = 24
    ACTION_STATE_VECTOR2F = 25
    ACTION_STATE_POSE = 27
    ACTION_SET_CREATE_INFO = 28
    ACTION_CREATE_INFO = 29
    INSTANCE_PROPERTIES = 32
    FRAME_WAIT_INFO = 33
    COMPOSITION_LAYER_PROJECTION = 35
    COMPOSITION_LAYER_QUAD = 36
    REFERENCE_SPACE_CREATE_INFO = 37
    ACTION_SPACE_CREATE_INFO = 38
    EVENT_DATA_REFERENCE_SPACE_CHANGE_PENDING = 40
    VIEW_CONFIGURATION_VIEW = 41
    SPACE_LOCATION = 42
    SPACE_VELOCITY = 43
    FRAME_STATE = 44
    VIEW_CONFIGURATION_PROPERTIES = 45
    FRAME_BEGIN_INFO = 46
    COMPOSITION_LAYER_PROJECTION_VIEW = 48
    EVENT_DATA_EVENTS_LOST = 49
    INTERACTION_PROFILE_SUGGESTED_BINDING = 51
    EVENT_DATA_INTERACTION_PROFILE_CHANGED = 52
    INTERACTION_PROFILE_STATE = 53
    SWAPCHAIN_IMAGE_ACQUIRE_INFO = 55
    SWAPCHAIN_IMAGE_WAIT_INFO = 56
    SWAPCHAIN_IMAGE_RELEASE_INFO = 57
    ACTION_STATE_GET_INFO = 58
    HAPTIC_ACTION_INFO = 59
    SESSION_ACTION_SETS_ATTACH_INFO = 60
    ACTIONS_SYNC_INFO = 61
    BOUND_SOURCES_FOR_ACTION_ENUMERATE_INFO = 62
    INPUT_SOURCE_LOCALIZED_NAME_GET_INFO = 63
    COMPOSITION_LAYER_CUBE_KHR = 1000006000
    INSTANCE_CREATE_INFO_ANDROID_KHR = 1000008000
    COMPOSITION_LAYER_DEPTH_INFO_KHR = 1000010000
    VULKAN_SWAPCHAIN_FORMAT_LIST_CREATE_INFO_KHR = 1000014000
    EVENT_DATA_PERF_SETTINGS_EXT = 1000015000
    COMPOSITION_LAYER_CYLINDER_KHR = 1000017000
    COMPOSITION_LAYER_EQUIRECT_KHR = 1000018000
    DEBUG_UTILS_OBJECT_NAME_INFO_EXT = 1000019000
    DEBUG_UTILS_MESSENGER_CALLBACK_DATA_EXT = 1000019001
    DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT = 1000019002
    DEBUG_UTILS_LABEL_EXT = 1000019003
    GRAPHICS_BINDING_OPENGL_WIN32_KHR = 1000023000
    GRAPHICS_BINDING_OPENGL_XLIB_KHR = 1000023001
    GRAPHICS_BINDING_OPENGL_XCB_KHR = 1000023002
    GRAPHICS_BINDING_OPENGL_WAYLAND_KHR = 1000023003
    SWAPCHAIN_IMAGE_OPENGL_KHR = 1000023004
    GRAPHICS_REQUIREMENTS_OPENGL_KHR = 1000023005
    GRAPHICS_BINDING_OPENGL_ES_ANDROID_KHR = 1000024001
    SWAPCHAIN_IMAGE_OPENGL_ES_KHR = 1000024002
    GRAPHICS_REQUIREMENTS_OPENGL_ES_KHR = 1000024003
    GRAPHICS_BINDING_VULKAN_KHR = 1000025000
    SWAPCHAIN_IMAGE_VULKAN_KHR = 1000025001
    GRAPHICS_REQUIREMENTS_VULKAN_KHR = 1000025002
    GRAPHICS_BINDING_D3D11_KHR = 1000027000
    SWAPCHAIN_IMAGE_D3D11_KHR = 1000027001
    GRAPHICS_REQUIREMENTS_D3D11_KHR = 1000027002
    GRAPHICS_BINDING_D3D12_KHR = 1000028000
    SWAPCHAIN_IMAGE_D3D12_KHR = 1000028001
    GRAPHICS_REQUIREMENTS_D3D12_KHR = 1000028002
    SYSTEM_EYE_GAZE_INTERACTION_PROPERTIES_EXT = 1000030000
    EYE_GAZE_SAMPLE_TIME_EXT = 1000030001
    VISIBILITY_MASK_KHR = 1000031000
    EVENT_DATA_VISIBILITY_MASK_CHANGED_KHR = 1000031001
    SESSION_CREATE_INFO_OVERLAY_EXTX = 1000033000
    EVENT_DATA_MAIN_SESSION_VISIBILITY_CHANGED_EXTX = 1000033003
    COMPOSITION_LAYER_COLOR_SCALE_BIAS_KHR = 1000034000
    SPATIAL_ANCHOR_CREATE_INFO_MSFT = 1000039000
    SPATIAL_ANCHOR_SPACE_CREATE_INFO_MSFT = 1000039001
    COMPOSITION_LAYER_IMAGE_LAYOUT_FB = 1000040000
    COMPOSITION_LAYER_ALPHA_BLEND_FB = 1000041001
    VIEW_CONFIGURATION_DEPTH_RANGE_EXT = 1000046000
    GRAPHICS_BINDING_EGL_MNDX = 1000048004
    SPATIAL_GRAPH_NODE_SPACE_CREATE_INFO_MSFT = 1000049000
    SPATIAL_GRAPH_STATIC_NODE_BINDING_CREATE_INFO_MSFT = 1000049001
    SPATIAL_GRAPH_NODE_BINDING_PROPERTIES_GET_INFO_MSFT = 1000049002
    SPATIAL_GRAPH_NODE_BINDING_PROPERTIES_MSFT = 1000049003
    SYSTEM_HAND_TRACKING_PROPERTIES_EXT = 1000051000
    HAND_TRACKER_CREATE_INFO_EXT = 1000051001
    HAND_JOINTS_LOCATE_INFO_EXT = 1000051002
    HAND_JOINT_LOCATIONS_EXT = 1000051003
    HAND_JOINT_VELOCITIES_EXT = 1000051004
    SYSTEM_HAND_TRACKING_MESH_PROPERTIES_MSFT = 1000052000
    HAND_MESH_SPACE_CREATE_INFO_MSFT = 1000052001
    HAND_MESH_UPDATE_INFO_MSFT = 1000052002
    HAND_MESH_MSFT = 1000052003
    HAND_POSE_TYPE_INFO_MSFT = 1000052004
    SECONDARY_VIEW_CONFIGURATION_SESSION_BEGIN_INFO_MSFT = 1000053000
    SECONDARY_VIEW_CONFIGURATION_STATE_MSFT = 1000053001
    SECONDARY_VIEW_CONFIGURATION_FRAME_STATE_MSFT = 1000053002
    SECONDARY_VIEW_CONFIGURATION_FRAME_END_INFO_MSFT = 1000053003
    SECONDARY_VIEW_CONFIGURATION_LAYER_INFO_MSFT = 1000053004
    SECONDARY_VIEW_CONFIGURATION_SWAPCHAIN_CREATE_INFO_MSFT = 1000053005
    CONTROLLER_MODEL_KEY_STATE_MSFT = 1000055000
    CONTROLLER_MODEL_NODE_PROPERTIES_MSFT = 1000055001
    CONTROLLER_MODEL_PROPERTIES_MSFT = 1000055002
    CONTROLLER_MODEL_NODE_STATE_MSFT = 1000055003
    CONTROLLER_MODEL_STATE_MSFT = 1000055004
    VIEW_CONFIGURATION_VIEW_FOV_EPIC = 1000059000
    HOLOGRAPHIC_WINDOW_ATTACHMENT_MSFT = 1000063000
    COMPOSITION_LAYER_REPROJECTION_INFO_MSFT = 1000066000
    COMPOSITION_LAYER_REPROJECTION_PLANE_OVERRIDE_MSFT = 1000066001
    ANDROID_SURFACE_SWAPCHAIN_CREATE_INFO_FB = 1000070000
    COMPOSITION_LAYER_SECURE_CONTENT_FB = 1000072000
    INTERACTION_PROFILE_DPAD_BINDING_EXT = 1000078000
    INTERACTION_PROFILE_ANALOG_THRESHOLD_VALVE = 1000079000
    HAND_JOINTS_MOTION_RANGE_INFO_EXT = 1000080000
    LOADER_INIT_INFO_ANDROID_KHR = 1000089000
    VULKAN_INSTANCE_CREATE_INFO_KHR = 1000090000
    VULKAN_DEVICE_CREATE_INFO_KHR = 1000090001
    VULKAN_GRAPHICS_DEVICE_GET_INFO_KHR = 1000090003
    COMPOSITION_LAYER_EQUIRECT2_KHR = 1000091000
    SCENE_OBSERVER_CREATE_INFO_MSFT = 1000097000
    SCENE_CREATE_INFO_MSFT = 1000097001
    NEW_SCENE_COMPUTE_INFO_MSFT = 1000097002
    VISUAL_MESH_COMPUTE_LOD_INFO_MSFT = 1000097003
    SCENE_COMPONENTS_MSFT = 1000097004
    SCENE_COMPONENTS_GET_INFO_MSFT = 1000097005
    SCENE_COMPONENT_LOCATIONS_MSFT = 1000097006
    SCENE_COMPONENTS_LOCATE_INFO_MSFT = 1000097007
    SCENE_OBJECTS_MSFT = 1000097008
    SCENE_COMPONENT_PARENT_FILTER_INFO_MSFT = 1000097009
    SCENE_OBJECT_TYPES_FILTER_INFO_MSFT = 1000097010
    SCENE_PLANES_MSFT = 1000097011
    SCENE_PLANE_ALIGNMENT_FILTER_INFO_MSFT = 1000097012
    SCENE_MESHES_MSFT = 1000097013
    SCENE_MESH_BUFFERS_GET_INFO_MSFT = 1000097014
    SCENE_MESH_BUFFERS_MSFT = 1000097015
    SCENE_MESH_VERTEX_BUFFER_MSFT = 1000097016
    SCENE_MESH_INDICES_UINT32_MSFT = 1000097017
    SCENE_MESH_INDICES_UINT16_MSFT = 1000097018
    SERIALIZED_SCENE_FRAGMENT_DATA_GET_INFO_MSFT = 1000098000
    SCENE_DESERIALIZE_INFO_MSFT = 1000098001
    EVENT_DATA_DISPLAY_REFRESH_RATE_CHANGED_FB = 1000101000
    VIVE_TRACKER_PATHS_HTCX = 1000103000
    EVENT_DATA_VIVE_TRACKER_CONNECTED_HTCX = 1000103001
    SYSTEM_FACIAL_TRACKING_PROPERTIES_HTC = 1000104000
    FACIAL_TRACKER_CREATE_INFO_HTC = 1000104001
    FACIAL_EXPRESSIONS_HTC = 1000104002
    SYSTEM_COLOR_SPACE_PROPERTIES_FB = 1000108000
    HAND_TRACKING_MESH_FB = 1000110001
    HAND_TRACKING_SCALE_FB = 1000110003
    HAND_TRACKING_AIM_STATE_FB = 1000111001
    HAND_TRACKING_CAPSULES_STATE_FB = 1000112000
    SYSTEM_SPATIAL_ENTITY_PROPERTIES_FB = 1000113004
    SPATIAL_ANCHOR_CREATE_INFO_FB = 1000113003
    SPACE_COMPONENT_STATUS_SET_INFO_FB = 1000113007
    SPACE_COMPONENT_STATUS_FB = 1000113001
    EVENT_DATA_SPATIAL_ANCHOR_CREATE_COMPLETE_FB = 1000113005
    EVENT_DATA_SPACE_SET_STATUS_COMPLETE_FB = 1000113006
    FOVEATION_PROFILE_CREATE_INFO_FB = 1000114000
    SWAPCHAIN_CREATE_INFO_FOVEATION_FB = 1000114001
    SWAPCHAIN_STATE_FOVEATION_FB = 1000114002
    FOVEATION_LEVEL_PROFILE_CREATE_INFO_FB = 1000115000
    KEYBOARD_SPACE_CREATE_INFO_FB = 1000116009
    KEYBOARD_TRACKING_QUERY_FB = 1000116004
    SYSTEM_KEYBOARD_TRACKING_PROPERTIES_FB = 1000116002
    TRIANGLE_MESH_CREATE_INFO_FB = 1000117001
    SYSTEM_PASSTHROUGH_PROPERTIES_FB = 1000118000
    PASSTHROUGH_CREATE_INFO_FB = 1000118001
    PASSTHROUGH_LAYER_CREATE_INFO_FB = 1000118002
    COMPOSITION_LAYER_PASSTHROUGH_FB = 1000118003
    GEOMETRY_INSTANCE_CREATE_INFO_FB = 1000118004
    GEOMETRY_INSTANCE_TRANSFORM_FB = 1000118005
    SYSTEM_PASSTHROUGH_PROPERTIES2_FB = 1000118006
    PASSTHROUGH_STYLE_FB = 1000118020
    PASSTHROUGH_COLOR_MAP_MONO_TO_RGBA_FB = 1000118021
    PASSTHROUGH_COLOR_MAP_MONO_TO_MONO_FB = 1000118022
    PASSTHROUGH_BRIGHTNESS_CONTRAST_SATURATION_FB = 1000118023
    EVENT_DATA_PASSTHROUGH_STATE_CHANGED_FB = 1000118030
    RENDER_MODEL_PATH_INFO_FB = 1000119000
    RENDER_MODEL_PROPERTIES_FB = 1000119001
    RENDER_MODEL_BUFFER_FB = 1000119002
    RENDER_MODEL_LOAD_INFO_FB = 1000119003
    SYSTEM_RENDER_MODEL_PROPERTIES_FB = 1000119004
    RENDER_MODEL_CAPABILITIES_REQUEST_FB = 1000119005
    BINDING_MODIFICATIONS_KHR = 1000120000
    VIEW_LOCATE_FOVEATED_RENDERING_VARJO = 1000121000
    FOVEATED_VIEW_CONFIGURATION_VIEW_VARJO = 1000121001
    SYSTEM_FOVEATED_RENDERING_PROPERTIES_VARJO = 1000121002
    COMPOSITION_LAYER_DEPTH_TEST_VARJO = 1000122000
    SYSTEM_MARKER_TRACKING_PROPERTIES_VARJO = 1000124000
    EVENT_DATA_MARKER_TRACKING_UPDATE_VARJO = 1000124001
    MARKER_SPACE_CREATE_INFO_VARJO = 1000124002
    SPATIAL_ANCHOR_PERSISTENCE_INFO_MSFT = 1000142000
    SPATIAL_ANCHOR_FROM_PERSISTED_ANCHOR_CREATE_INFO_MSFT = 1000142001
    SPACE_QUERY_INFO_FB = 1000156001
    SPACE_QUERY_RESULTS_FB = 1000156002
    SPACE_STORAGE_LOCATION_FILTER_INFO_FB = 1000156003
    SPACE_UUID_FILTER_INFO_FB = 1000156054
    SPACE_COMPONENT_FILTER_INFO_FB = 1000156052
    EVENT_DATA_SPACE_QUERY_RESULTS_AVAILABLE_FB = 1000156103
    EVENT_DATA_SPACE_QUERY_COMPLETE_FB = 1000156104
    SPACE_SAVE_INFO_FB = 1000158000
    SPACE_ERASE_INFO_FB = 1000158001
    EVENT_DATA_SPACE_SAVE_COMPLETE_FB = 1000158106
    EVENT_DATA_SPACE_ERASE_COMPLETE_FB = 1000158107
    SWAPCHAIN_IMAGE_FOVEATION_VULKAN_FB = 1000160000
    SWAPCHAIN_STATE_ANDROID_SURFACE_DIMENSIONS_FB = 1000161000
    SWAPCHAIN_STATE_SAMPLER_OPENGL_ES_FB = 1000162000
    SWAPCHAIN_STATE_SAMPLER_VULKAN_FB = 1000163000
    COMPOSITION_LAYER_SPACE_WARP_INFO_FB = 1000171000
    SYSTEM_SPACE_WARP_PROPERTIES_FB = 1000171001
    SEMANTIC_LABELS_FB = 1000175000
    ROOM_LAYOUT_FB = 1000175001
    BOUNDARY_2D_FB = 1000175002
    DIGITAL_LENS_CONTROL_ALMALENCE = 1000196000
    SPACE_CONTAINER_FB = 1000199000
    PASSTHROUGH_KEYBOARD_HANDS_INTENSITY_FB = 1000203002
    COMPOSITION_LAYER_SETTINGS_FB = 1000204000
    VULKAN_SWAPCHAIN_CREATE_INFO_META = 1000227000
    PERFORMANCE_METRICS_STATE_META = 1000232001
    PERFORMANCE_METRICS_COUNTER_META = 1000232002
    GRAPHICS_BINDING_VULKAN2_KHR = 1000025000
    SWAPCHAIN_IMAGE_VULKAN2_KHR = 1000025001
    GRAPHICS_REQUIREMENTS_VULKAN2_KHR = 1000025002


class FormFactor(EnumBase):
    HEAD_MOUNTED_DISPLAY = 1
    HANDHELD_DISPLAY = 2


class ViewConfigurationType(EnumBase):
    PRIMARY_MONO = 1
    PRIMARY_STEREO = 2
    PRIMARY_QUAD_VARJO = 1000037000
    SECONDARY_MONO_FIRST_PERSON_OBSERVER_MSFT = 1000054000


class EnvironmentBlendMode(EnumBase):
    OPAQUE = 1
    ADDITIVE = 2
    ALPHA_BLEND = 3


class ReferenceSpaceType(EnumBase):
    VIEW = 1
    LOCAL = 2
    STAGE = 3
    UNBOUNDED_MSFT = 1000038000
    COMBINED_EYE_VARJO = 1000121000


class ActionType(EnumBase):
    BOOLEAN_INPUT = 1
    FLOAT_INPUT = 2
    VECTOR2F_INPUT = 3
    POSE_INPUT = 4
    VIBRATION_OUTPUT = 100


class EyeVisibility(EnumBase):
    BOTH = 0
    LEFT = 1
    RIGHT = 2


class SessionState(EnumBase):
    UNKNOWN = 0
    IDLE = 1
    READY = 2
    SYNCHRONIZED = 3
    VISIBLE = 4
    FOCUSED = 5
    STOPPING = 6
    LOSS_PENDING = 7
    EXITING = 8


class ObjectType(EnumBase):
    UNKNOWN = 0
    INSTANCE = 1
    SESSION = 2
    SWAPCHAIN = 3
    SPACE = 4
    ACTION_SET = 5
    ACTION = 6
    DEBUG_UTILS_MESSENGER_EXT = 1000019000
    SPATIAL_ANCHOR_MSFT = 1000039000
    SPATIAL_GRAPH_NODE_BINDING_MSFT = 1000049000
    HAND_TRACKER_EXT = 1000051000
    SCENE_OBSERVER_MSFT = 1000097000
    SCENE_MSFT = 1000097001
    FACIAL_TRACKER_HTC = 1000104000
    FOVEATION_PROFILE_FB = 1000114000
    TRIANGLE_MESH_FB = 1000117000
    PASSTHROUGH_FB = 1000118000
    PASSTHROUGH_LAYER_FB = 1000118002
    GEOMETRY_INSTANCE_FB = 1000118004
    SPATIAL_ANCHOR_STORE_CONNECTION_MSFT = 1000142000


class VisibilityMaskTypeKHR(EnumBase):
    HIDDEN_TRIANGLE_MESH = 1
    VISIBLE_TRIANGLE_MESH = 2
    LINE_LOOP = 3


class PerfSettingsDomainEXT(EnumBase):
    CPU = 1
    GPU = 2


class PerfSettingsSubDomainEXT(EnumBase):
    COMPOSITING = 1
    RENDERING = 2
    THERMAL = 3


class PerfSettingsLevelEXT(EnumBase):
    POWER_SAVINGS = 0
    SUSTAINED_LOW = 25
    SUSTAINED_HIGH = 50
    BOOST = 75


class PerfSettingsNotificationLevelEXT(EnumBase):
    NORMAL = 0
    WARNING = 25
    IMPAIRED = 75


class BlendFactorFB(EnumBase):
    ZERO = 0
    ONE = 1
    SRC_ALPHA = 2
    ONE_MINUS_SRC_ALPHA = 3
    DST_ALPHA = 4
    ONE_MINUS_DST_ALPHA = 5


class SpatialGraphNodeTypeMSFT(EnumBase):
    STATIC = 1
    DYNAMIC = 2


class HandEXT(EnumBase):
    LEFT = 1
    RIGHT = 2


class HandJointEXT(EnumBase):
    PALM = 0
    WRIST = 1
    THUMB_METACARPAL = 2
    THUMB_PROXIMAL = 3
    THUMB_DISTAL = 4
    THUMB_TIP = 5
    INDEX_METACARPAL = 6
    INDEX_PROXIMAL = 7
    INDEX_INTERMEDIATE = 8
    INDEX_DISTAL = 9
    INDEX_TIP = 10
    MIDDLE_METACARPAL = 11
    MIDDLE_PROXIMAL = 12
    MIDDLE_INTERMEDIATE = 13
    MIDDLE_DISTAL = 14
    MIDDLE_TIP = 15
    RING_METACARPAL = 16
    RING_PROXIMAL = 17
    RING_INTERMEDIATE = 18
    RING_DISTAL = 19
    RING_TIP = 20
    LITTLE_METACARPAL = 21
    LITTLE_PROXIMAL = 22
    LITTLE_INTERMEDIATE = 23
    LITTLE_DISTAL = 24
    LITTLE_TIP = 25


class HandJointSetEXT(EnumBase):
    DEFAULT = 0
    HAND_WITH_FOREARM_ULTRA = 1000149000


class HandPoseTypeMSFT(EnumBase):
    TRACKED = 0
    REFERENCE_OPEN_PALM = 1


class ReprojectionModeMSFT(EnumBase):
    DEPTH = 1
    PLANAR_FROM_DEPTH = 2
    PLANAR_MANUAL = 3
    ORIENTATION_ONLY = 4


class HandJointsMotionRangeEXT(EnumBase):
    UNOBSTRUCTED = 1
    CONFORMING_TO_CONTROLLER = 2


class SceneComputeFeatureMSFT(EnumBase):
    PLANE = 1
    PLANE_MESH = 2
    VISUAL_MESH = 3
    COLLIDER_MESH = 4
    SERIALIZE_SCENE = 1000098000


class SceneComputeConsistencyMSFT(EnumBase):
    SNAPSHOT_COMPLETE = 1
    SNAPSHOT_INCOMPLETE_FAST = 2
    OCCLUSION_OPTIMIZED = 3


class MeshComputeLodMSFT(EnumBase):
    COARSE = 1
    MEDIUM = 2
    FINE = 3
    UNLIMITED = 4


class SceneComponentTypeMSFT(EnumBase):
    INVALID = -1
    OBJECT = 1
    PLANE = 2
    VISUAL_MESH = 3
    COLLIDER_MESH = 4
    SERIALIZED_SCENE_FRAGMENT = 1000098000


class SceneObjectTypeMSFT(EnumBase):
    UNCATEGORIZED = -1
    BACKGROUND = 1
    WALL = 2
    FLOOR = 3
    CEILING = 4
    PLATFORM = 5
    INFERRED = 6


class ScenePlaneAlignmentTypeMSFT(EnumBase):
    NON_ORTHOGONAL = 0
    HORIZONTAL = 1
    VERTICAL = 2


class SceneComputeStateMSFT(EnumBase):
    NONE = 0
    UPDATING = 1
    COMPLETED = 2
    COMPLETED_WITH_ERROR = 3


class EyeExpressionHTC(EnumBase):
    LEFT_BLINK = 0
    LEFT_WIDE = 1
    RIGHT_BLINK = 2
    RIGHT_WIDE = 3
    LEFT_SQUEEZE = 4
    RIGHT_SQUEEZE = 5
    LEFT_DOWN = 6
    RIGHT_DOWN = 7
    LEFT_OUT = 8
    RIGHT_IN = 9
    LEFT_IN = 10
    RIGHT_OUT = 11
    LEFT_UP = 12
    RIGHT_UP = 13


class LipExpressionHTC(EnumBase):
    JAW_RIGHT = 0
    JAW_LEFT = 1
    JAW_FORWARD = 2
    JAW_OPEN = 3
    MOUTH_APE_SHAPE = 4
    MOUTH_UPPER_RIGHT = 5
    MOUTH_UPPER_LEFT = 6
    MOUTH_LOWER_RIGHT = 7
    MOUTH_LOWER_LEFT = 8
    MOUTH_UPPER_OVERTURN = 9
    MOUTH_LOWER_OVERTURN = 10
    MOUTH_POUT = 11
    MOUTH_SMILE_RIGHT = 12
    MOUTH_SMILE_LEFT = 13
    MOUTH_SAD_RIGHT = 14
    MOUTH_SAD_LEFT = 15
    CHEEK_PUFF_RIGHT = 16
    CHEEK_PUFF_LEFT = 17
    CHEEK_SUCK = 18
    MOUTH_UPPER_UPRIGHT = 19
    MOUTH_UPPER_UPLEFT = 20
    MOUTH_LOWER_DOWNRIGHT = 21
    MOUTH_LOWER_DOWNLEFT = 22
    MOUTH_UPPER_INSIDE = 23
    MOUTH_LOWER_INSIDE = 24
    MOUTH_LOWER_OVERLAY = 25
    TONGUE_LONGSTEP1 = 26
    TONGUE_LEFT = 27
    TONGUE_RIGHT = 28
    TONGUE_UP = 29
    TONGUE_DOWN = 30
    TONGUE_ROLL = 31
    TONGUE_LONGSTEP2 = 32
    TONGUE_UPRIGHT_MORPH = 33
    TONGUE_UPLEFT_MORPH = 34
    TONGUE_DOWNRIGHT_MORPH = 35
    TONGUE_DOWNLEFT_MORPH = 36


class FacialTrackingTypeHTC(EnumBase):
    EYE_DEFAULT = 1
    LIP_DEFAULT = 2


class ColorSpaceFB(EnumBase):
    UNMANAGED = 0
    REC2020 = 1
    REC709 = 2
    RIFT_CV1 = 3
    RIFT_S = 4
    QUEST = 5
    P3 = 6
    ADOBE_RGB = 7


class SpaceComponentTypeFB(EnumBase):
    LOCATABLE = 0
    STORABLE = 1
    BOUNDED_2D = 3
    BOUNDED_3D = 4
    SEMANTIC_LABELS = 5
    ROOM_LAYOUT = 6
    SPACE_CONTAINER = 7


class FoveationLevelFB(EnumBase):
    NONE = 0
    LOW = 1
    MEDIUM = 2
    HIGH = 3


class FoveationDynamicFB(EnumBase):
    DISABLED = 0
    LEVEL_ENABLED = 1


class WindingOrderFB(EnumBase):
    UNKNOWN = 0
    CW = 1
    CCW = 2


class PassthroughLayerPurposeFB(EnumBase):
    RECONSTRUCTION = 0
    PROJECTED = 1
    TRACKED_KEYBOARD_HANDS = 1000203001
    TRACKED_KEYBOARD_MASKED_HANDS = 1000203002


class HandForearmJointULTRALEAP(EnumBase):
    PALM = 0
    WRIST = 1
    THUMB_METACARPAL = 2
    THUMB_PROXIMAL = 3
    THUMB_DISTAL = 4
    THUMB_TIP = 5
    INDEX_METACARPAL = 6
    INDEX_PROXIMAL = 7
    INDEX_INTERMEDIATE = 8
    INDEX_DISTAL = 9
    INDEX_TIP = 10
    MIDDLE_METACARPAL = 11
    MIDDLE_PROXIMAL = 12
    MIDDLE_INTERMEDIATE = 13
    MIDDLE_DISTAL = 14
    MIDDLE_TIP = 15
    RING_METACARPAL = 16
    RING_PROXIMAL = 17
    RING_INTERMEDIATE = 18
    RING_DISTAL = 19
    RING_TIP = 20
    LITTLE_METACARPAL = 21
    LITTLE_PROXIMAL = 22
    LITTLE_INTERMEDIATE = 23
    LITTLE_DISTAL = 24
    LITTLE_TIP = 25
    ELBOW = 26


class SpaceQueryActionFB(EnumBase):
    LOAD = 0


class SpaceStorageLocationFB(EnumBase):
    INVALID = 0
    LOCAL = 1


class SpacePersistenceModeFB(EnumBase):
    INVALID = 0
    INDEFINITE = 1


class PerformanceMetricsCounterUnitMETA(EnumBase):
    GENERIC = 0
    PERCENTAGE = 1
    MILLISECONDS = 2
    BYTES = 3
    HERTZ = 4


class InstanceCreateFlags(FlagBase):
    NONE = 0x00000000


class SessionCreateFlags(FlagBase):
    NONE = 0x00000000


class SpaceVelocityFlags(FlagBase):
    NONE = 0x00000000
    LINEAR_VALID_BIT = 0x00000001
    ANGULAR_VALID_BIT = 0x00000002


class SpaceLocationFlags(FlagBase):
    NONE = 0x00000000
    ORIENTATION_VALID_BIT = 0x00000001
    POSITION_VALID_BIT = 0x00000002
    ORIENTATION_TRACKED_BIT = 0x00000004
    POSITION_TRACKED_BIT = 0x00000008


class SwapchainCreateFlags(FlagBase):
    NONE = 0x00000000
    PROTECTED_CONTENT_BIT = 0x00000001
    STATIC_IMAGE_BIT = 0x00000002


class SwapchainUsageFlags(FlagBase):
    NONE = 0x00000000
    COLOR_ATTACHMENT_BIT = 0x00000001
    DEPTH_STENCIL_ATTACHMENT_BIT = 0x00000002
    UNORDERED_ACCESS_BIT = 0x00000004
    TRANSFER_SRC_BIT = 0x00000008
    TRANSFER_DST_BIT = 0x00000010
    SAMPLED_BIT = 0x00000020
    MUTABLE_FORMAT_BIT = 0x00000040
    INPUT_ATTACHMENT_BIT_MND = 0x00000080
    INPUT_ATTACHMENT_BIT_KHR = 0x00000080


class CompositionLayerFlags(FlagBase):
    NONE = 0x00000000
    CORRECT_CHROMATIC_ABERRATION_BIT = 0x00000001
    BLEND_TEXTURE_SOURCE_ALPHA_BIT = 0x00000002
    UNPREMULTIPLIED_ALPHA_BIT = 0x00000004


class ViewStateFlags(FlagBase):
    NONE = 0x00000000
    ORIENTATION_VALID_BIT = 0x00000001
    POSITION_VALID_BIT = 0x00000002
    ORIENTATION_TRACKED_BIT = 0x00000004
    POSITION_TRACKED_BIT = 0x00000008


class InputSourceLocalizedNameFlags(FlagBase):
    NONE = 0x00000000
    USER_PATH_BIT = 0x00000001
    INTERACTION_PROFILE_BIT = 0x00000002
    COMPONENT_BIT = 0x00000004


class DebugUtilsMessageSeverityFlagsEXT(FlagBase):
    NONE = 0x00000000
    VERBOSE_BIT = 0x00000001
    INFO_BIT = 0x00000010
    WARNING_BIT = 0x00000100
    ERROR_BIT = 0x00001000


class DebugUtilsMessageTypeFlagsEXT(FlagBase):
    NONE = 0x00000000
    GENERAL_BIT = 0x00000001
    VALIDATION_BIT = 0x00000002
    PERFORMANCE_BIT = 0x00000004
    CONFORMANCE_BIT = 0x00000008


class OverlaySessionCreateFlagsEXTX(FlagBase):
    NONE = 0x00000000


class OverlayMainSessionFlagsEXTX(FlagBase):
    NONE = 0x00000000
    ENABLED_COMPOSITION_LAYER_INFO_DEPTH_BIT = 0x00000001


class CompositionLayerImageLayoutFlagsFB(FlagBase):
    NONE = 0x00000000
    VERTICAL_FLIP_BIT = 0x00000001


class CompositionLayerSecureContentFlagsFB(FlagBase):
    NONE = 0x00000000
    EXCLUDE_LAYER_BIT = 0x00000001
    REPLACE_LAYER_BIT = 0x00000002


class HandTrackingAimFlagsFB(FlagBase):
    NONE = 0x00000000
    COMPUTED_BIT = 0x00000001
    VALID_BIT = 0x00000002
    INDEX_PINCHING_BIT = 0x00000004
    MIDDLE_PINCHING_BIT = 0x00000008
    RING_PINCHING_BIT = 0x00000010
    LITTLE_PINCHING_BIT = 0x00000020
    SYSTEM_GESTURE_BIT = 0x00000040
    DOMINANT_HAND_BIT = 0x00000080
    MENU_PRESSED_BIT = 0x00000100


class SwapchainCreateFoveationFlagsFB(FlagBase):
    NONE = 0x00000000
    SCALED_BIN_BIT = 0x00000001
    FRAGMENT_DENSITY_MAP_BIT = 0x00000002


class SwapchainStateFoveationFlagsFB(FlagBase):
    NONE = 0x00000000


class KeyboardTrackingFlagsFB(FlagBase):
    NONE = 0x00000000
    EXISTS_BIT = 0x00000001
    LOCAL_BIT = 0x00000002
    REMOTE_BIT = 0x00000004
    CONNECTED_BIT = 0x00000008


class KeyboardTrackingQueryFlagsFB(FlagBase):
    NONE = 0x00000000
    LOCAL_BIT = 0x00000002
    REMOTE_BIT = 0x00000004


class TriangleMeshFlagsFB(FlagBase):
    NONE = 0x00000000
    MUTABLE_BIT = 0x00000001


class PassthroughCapabilityFlagsFB(FlagBase):
    NONE = 0x00000000
    BIT = 0x00000001
    COLOR_BIT = 0x00000002
    LAYER_DEPTH_BIT = 0x00000004


class PassthroughFlagsFB(FlagBase):
    NONE = 0x00000000
    IS_RUNNING_AT_CREATION_BIT = 0x00000001
    LAYER_DEPTH_BIT = 0x00000002


class PassthroughStateChangedFlagsFB(FlagBase):
    NONE = 0x00000000
    REINIT_REQUIRED_BIT = 0x00000001
    NON_RECOVERABLE_ERROR_BIT = 0x00000002
    RECOVERABLE_ERROR_BIT = 0x00000004
    RESTORED_ERROR_BIT = 0x00000008


class RenderModelFlagsFB(FlagBase):
    NONE = 0x00000000
    SUPPORTS_GLTF_2_0_SUBSET_1_BIT = 0x00000001
    SUPPORTS_GLTF_2_0_SUBSET_2_BIT = 0x00000002


class CompositionLayerSpaceWarpInfoFlagsFB(FlagBase):
    NONE = 0x00000000
    FRAME_SKIP_BIT = 0x00000001


class DigitalLensControlFlagsALMALENCE(FlagBase):
    NONE = 0x00000000
    PROCESSING_DISABLE_BIT = 0x00000001


class CompositionLayerSettingsFlagsFB(FlagBase):
    NONE = 0x00000000
    NORMAL_SUPER_SAMPLING_BIT = 0x00000001
    QUALITY_SUPER_SAMPLING_BIT = 0x00000002
    NORMAL_SHARPENING_BIT = 0x00000004
    QUALITY_SHARPENING_BIT = 0x00000008


class PerformanceMetricsCounterFlagsMETA(FlagBase):
    NONE = 0x00000000
    ANY_VALUE_VALID_BIT = 0x00000001
    UINT_VALUE_VALID_BIT = 0x00000002
    FLOAT_VALUE_VALID_BIT = 0x00000004


__all__ = [
    "ActionType",
    "BlendFactorFB",
    "ColorSpaceFB",
    "CompositionLayerFlags",
    "CompositionLayerImageLayoutFlagsFB",
    "CompositionLayerSecureContentFlagsFB",
    "CompositionLayerSettingsFlagsFB",
    "CompositionLayerSpaceWarpInfoFlagsFB",
    "DebugUtilsMessageSeverityFlagsEXT",
    "DebugUtilsMessageTypeFlagsEXT",
    "DigitalLensControlFlagsALMALENCE",
    "EnumBase",
    "EnvironmentBlendMode",
    "EyeExpressionHTC",
    "EyeVisibility",
    "FacialTrackingTypeHTC",
    "FlagBase",
    "FormFactor",
    "FoveationDynamicFB",
    "FoveationLevelFB",
    "HandEXT",
    "HandForearmJointULTRALEAP",
    "HandJointEXT",
    "HandJointSetEXT",
    "HandJointsMotionRangeEXT",
    "HandPoseTypeMSFT",
    "HandTrackingAimFlagsFB",
    "InputSourceLocalizedNameFlags",
    "InstanceCreateFlags",
    "KeyboardTrackingFlagsFB",
    "KeyboardTrackingQueryFlagsFB",
    "LipExpressionHTC",
    "MeshComputeLodMSFT",
    "ObjectType",
    "OverlayMainSessionFlagsEXTX",
    "OverlaySessionCreateFlagsEXTX",
    "PassthroughCapabilityFlagsFB",
    "PassthroughFlagsFB",
    "PassthroughLayerPurposeFB",
    "PassthroughStateChangedFlagsFB",
    "PerfSettingsDomainEXT",
    "PerfSettingsLevelEXT",
    "PerfSettingsNotificationLevelEXT",
    "PerfSettingsSubDomainEXT",
    "PerformanceMetricsCounterFlagsMETA",
    "PerformanceMetricsCounterUnitMETA",
    "ReferenceSpaceType",
    "RenderModelFlagsFB",
    "ReprojectionModeMSFT",
    "Result",
    "SceneComponentTypeMSFT",
    "SceneComputeConsistencyMSFT",
    "SceneComputeFeatureMSFT",
    "SceneComputeStateMSFT",
    "SceneObjectTypeMSFT",
    "ScenePlaneAlignmentTypeMSFT",
    "SessionCreateFlags",
    "SessionState",
    "SpaceComponentTypeFB",
    "SpaceLocationFlags",
    "SpacePersistenceModeFB",
    "SpaceQueryActionFB",
    "SpaceStorageLocationFB",
    "SpaceVelocityFlags",
    "SpatialGraphNodeTypeMSFT",
    "StructureType",
    "SwapchainCreateFlags",
    "SwapchainCreateFoveationFlagsFB",
    "SwapchainStateFoveationFlagsFB",
    "SwapchainUsageFlags",
    "TriangleMeshFlagsFB",
    "ViewConfigurationType",
    "ViewStateFlags",
    "VisibilityMaskTypeKHR",
    "WindingOrderFB",
    "failed",
    "succeeded",
    "unqualified_success",
]



# Warning: this file is auto-generated. Do not edit.

from ctypes import POINTER, c_char, c_char_p, c_int64, c_uint32

"""
File xr.raw_functions.py

Defines low-level ctypes function definitions for use by
higher-level pythonic functions in pyopenxr.
"""

from .library import openxr_loader_library
from .enums import *
from .typedefs import *


# ctypes Function definitions

xrGetInstanceProcAddr = openxr_loader_library.xrGetInstanceProcAddr
xrGetInstanceProcAddr.__doc__ = """
    Gets a function pointer for an OpenXR function.
"""
xrGetInstanceProcAddr.restype = Result
xrGetInstanceProcAddr.argtypes = [
    Instance,  # instance
    c_char_p,  # name
    POINTER(PFN_xrVoidFunction),  # function
]

xrEnumerateApiLayerProperties = openxr_loader_library.xrEnumerateApiLayerProperties
xrEnumerateApiLayerProperties.__doc__ = """
    Returns up to requested number of global layer properties.
"""
xrEnumerateApiLayerProperties.restype = Result
xrEnumerateApiLayerProperties.argtypes = [
    c_uint32,  # property_capacity_input
    POINTER(c_uint32),  # property_count_output
    POINTER(ApiLayerProperties),  # properties
]

xrEnumerateInstanceExtensionProperties = openxr_loader_library.xrEnumerateInstanceExtensionProperties
xrEnumerateInstanceExtensionProperties.__doc__ = """
    Returns properties of available instance extensions.
"""
xrEnumerateInstanceExtensionProperties.restype = Result
xrEnumerateInstanceExtensionProperties.argtypes = [
    c_char_p,  # layer_name
    c_uint32,  # property_capacity_input
    POINTER(c_uint32),  # property_count_output
    POINTER(ExtensionProperties),  # properties
]

xrCreateInstance = openxr_loader_library.xrCreateInstance
xrCreateInstance.__doc__ = """
    Creates an OpenXR Instance.
"""
xrCreateInstance.restype = Result
xrCreateInstance.argtypes = [
    POINTER(InstanceCreateInfo),  # create_info
    POINTER(Instance),  # instance
]

xrDestroyInstance = openxr_loader_library.xrDestroyInstance
xrDestroyInstance.__doc__ = """
    Destroy an instance of OpenXR.
"""
xrDestroyInstance.restype = Result
xrDestroyInstance.argtypes = [
    Instance,  # instance
]

xrGetInstanceProperties = openxr_loader_library.xrGetInstanceProperties
xrGetInstanceProperties.__doc__ = """
    Gets information about the instance.
"""
xrGetInstanceProperties.restype = Result
xrGetInstanceProperties.argtypes = [
    Instance,  # instance
    POINTER(InstanceProperties),  # instance_properties
]

xrPollEvent = openxr_loader_library.xrPollEvent
xrPollEvent.__doc__ = """
    Polls for events.
"""
xrPollEvent.restype = Result
xrPollEvent.argtypes = [
    Instance,  # instance
    POINTER(EventDataBuffer),  # event_data
]

xrResultToString = openxr_loader_library.xrResultToString
xrResultToString.__doc__ = """
    Converts an XrResult to a UTF-8 string.
"""
xrResultToString.restype = Result
xrResultToString.argtypes = [
    Instance,  # instance
    Result.ctype(),  # value
    (c_char * 64),  # buffer
]

xrStructureTypeToString = openxr_loader_library.xrStructureTypeToString
xrStructureTypeToString.__doc__ = """
    Converts an XrStructureType to a UTF-8 string.
"""
xrStructureTypeToString.restype = Result
xrStructureTypeToString.argtypes = [
    Instance,  # instance
    StructureType.ctype(),  # value
    (c_char * 64),  # buffer
]

xrGetSystem = openxr_loader_library.xrGetSystem
xrGetSystem.__doc__ = """
    Gets a system identifier.
"""
xrGetSystem.restype = Result
xrGetSystem.argtypes = [
    Instance,  # instance
    POINTER(SystemGetInfo),  # get_info
    POINTER(SystemId),  # system_id
]

xrGetSystemProperties = openxr_loader_library.xrGetSystemProperties
xrGetSystemProperties.__doc__ = """
    Gets the properties of a particular system.
"""
xrGetSystemProperties.restype = Result
xrGetSystemProperties.argtypes = [
    Instance,  # instance
    SystemId,  # system_id
    POINTER(SystemProperties),  # properties
]

xrEnumerateEnvironmentBlendModes = openxr_loader_library.xrEnumerateEnvironmentBlendModes
xrEnumerateEnvironmentBlendModes.__doc__ = """
    Lists environment blend modes.
"""
xrEnumerateEnvironmentBlendModes.restype = Result
xrEnumerateEnvironmentBlendModes.argtypes = [
    Instance,  # instance
    SystemId,  # system_id
    ViewConfigurationType.ctype(),  # view_configuration_type
    c_uint32,  # environment_blend_mode_capacity_input
    POINTER(c_uint32),  # environment_blend_mode_count_output
    POINTER(EnvironmentBlendMode.ctype()),  # environment_blend_modes
]

xrCreateSession = openxr_loader_library.xrCreateSession
xrCreateSession.__doc__ = """
    Creates an XrSession.
"""
xrCreateSession.restype = Result
xrCreateSession.argtypes = [
    Instance,  # instance
    POINTER(SessionCreateInfo),  # create_info
    POINTER(Session),  # session
]

xrDestroySession = openxr_loader_library.xrDestroySession
xrDestroySession.__doc__ = """
    Destroys an XrSession.
"""
xrDestroySession.restype = Result
xrDestroySession.argtypes = [
    Session,  # session
]

xrEnumerateReferenceSpaces = openxr_loader_library.xrEnumerateReferenceSpaces
xrEnumerateReferenceSpaces.__doc__ = """
    Enumerate available reference spaces.
"""
xrEnumerateReferenceSpaces.restype = Result
xrEnumerateReferenceSpaces.argtypes = [
    Session,  # session
    c_uint32,  # space_capacity_input
    POINTER(c_uint32),  # space_count_output
    POINTER(ReferenceSpaceType.ctype()),  # spaces
]

xrCreateReferenceSpace = openxr_loader_library.xrCreateReferenceSpace
xrCreateReferenceSpace.__doc__ = """
    Creates a reference space.
"""
xrCreateReferenceSpace.restype = Result
xrCreateReferenceSpace.argtypes = [
    Session,  # session
    POINTER(ReferenceSpaceCreateInfo),  # create_info
    POINTER(Space),  # space
]

xrGetReferenceSpaceBoundsRect = openxr_loader_library.xrGetReferenceSpaceBoundsRect
xrGetReferenceSpaceBoundsRect.__doc__ = """
    Gets the bounds rectangle of a reference space.
"""
xrGetReferenceSpaceBoundsRect.restype = Result
xrGetReferenceSpaceBoundsRect.argtypes = [
    Session,  # session
    ReferenceSpaceType.ctype(),  # reference_space_type
    POINTER(Extent2Df),  # bounds
]

xrCreateActionSpace = openxr_loader_library.xrCreateActionSpace
xrCreateActionSpace.__doc__ = """
    Creates a space based on a pose action.
"""
xrCreateActionSpace.restype = Result
xrCreateActionSpace.argtypes = [
    Session,  # session
    POINTER(ActionSpaceCreateInfo),  # create_info
    POINTER(Space),  # space
]

xrLocateSpace = openxr_loader_library.xrLocateSpace
xrLocateSpace.__doc__ = """
    Locates a space with reference to another space.
"""
xrLocateSpace.restype = Result
xrLocateSpace.argtypes = [
    Space,  # space
    Space,  # base_space
    Time,  # time
    POINTER(SpaceLocation),  # location
]

xrDestroySpace = openxr_loader_library.xrDestroySpace
xrDestroySpace.__doc__ = """
    Creates a space based on a pose action.
"""
xrDestroySpace.restype = Result
xrDestroySpace.argtypes = [
    Space,  # space
]

xrEnumerateViewConfigurations = openxr_loader_library.xrEnumerateViewConfigurations
xrEnumerateViewConfigurations.__doc__ = """
    Enumerates supported view configurations.
"""
xrEnumerateViewConfigurations.restype = Result
xrEnumerateViewConfigurations.argtypes = [
    Instance,  # instance
    SystemId,  # system_id
    c_uint32,  # view_configuration_type_capacity_input
    POINTER(c_uint32),  # view_configuration_type_count_output
    POINTER(ViewConfigurationType.ctype()),  # view_configuration_types
]

xrGetViewConfigurationProperties = openxr_loader_library.xrGetViewConfigurationProperties
xrGetViewConfigurationProperties.__doc__ = """
    Gets information for a view configuration.
"""
xrGetViewConfigurationProperties.restype = Result
xrGetViewConfigurationProperties.argtypes = [
    Instance,  # instance
    SystemId,  # system_id
    ViewConfigurationType.ctype(),  # view_configuration_type
    POINTER(ViewConfigurationProperties),  # configuration_properties
]

xrEnumerateViewConfigurationViews = openxr_loader_library.xrEnumerateViewConfigurationViews
xrEnumerateViewConfigurationViews.__doc__ = """
    Gets view configuration views.
"""
xrEnumerateViewConfigurationViews.restype = Result
xrEnumerateViewConfigurationViews.argtypes = [
    Instance,  # instance
    SystemId,  # system_id
    ViewConfigurationType.ctype(),  # view_configuration_type
    c_uint32,  # view_capacity_input
    POINTER(c_uint32),  # view_count_output
    POINTER(ViewConfigurationView),  # views
]

xrEnumerateSwapchainFormats = openxr_loader_library.xrEnumerateSwapchainFormats
xrEnumerateSwapchainFormats.__doc__ = """
    Enumerates swapchain formats.
"""
xrEnumerateSwapchainFormats.restype = Result
xrEnumerateSwapchainFormats.argtypes = [
    Session,  # session
    c_uint32,  # format_capacity_input
    POINTER(c_uint32),  # format_count_output
    POINTER(c_int64),  # formats
]

xrCreateSwapchain = openxr_loader_library.xrCreateSwapchain
xrCreateSwapchain.__doc__ = """
    Creates an XrSwapchain.
"""
xrCreateSwapchain.restype = Result
xrCreateSwapchain.argtypes = [
    Session,  # session
    POINTER(SwapchainCreateInfo),  # create_info
    POINTER(Swapchain),  # swapchain
]

xrDestroySwapchain = openxr_loader_library.xrDestroySwapchain
xrDestroySwapchain.__doc__ = """
    Destroys an XrSwapchain.
"""
xrDestroySwapchain.restype = Result
xrDestroySwapchain.argtypes = [
    Swapchain,  # swapchain
]

xrEnumerateSwapchainImages = openxr_loader_library.xrEnumerateSwapchainImages
xrEnumerateSwapchainImages.__doc__ = """
    Gets images from an XrSwapchain.
"""
xrEnumerateSwapchainImages.restype = Result
xrEnumerateSwapchainImages.argtypes = [
    Swapchain,  # swapchain
    c_uint32,  # image_capacity_input
    POINTER(c_uint32),  # image_count_output
    POINTER(SwapchainImageBaseHeader),  # images
]

xrAcquireSwapchainImage = openxr_loader_library.xrAcquireSwapchainImage
xrAcquireSwapchainImage.__doc__ = """
    Acquire a swapchain image.
"""
xrAcquireSwapchainImage.restype = Result
xrAcquireSwapchainImage.argtypes = [
    Swapchain,  # swapchain
    POINTER(SwapchainImageAcquireInfo),  # acquire_info
    POINTER(c_uint32),  # index
]

xrWaitSwapchainImage = openxr_loader_library.xrWaitSwapchainImage
xrWaitSwapchainImage.__doc__ = """
    Wait for a swapchain image to be available.
"""
xrWaitSwapchainImage.restype = Result
xrWaitSwapchainImage.argtypes = [
    Swapchain,  # swapchain
    POINTER(SwapchainImageWaitInfo),  # wait_info
]

xrReleaseSwapchainImage = openxr_loader_library.xrReleaseSwapchainImage
xrReleaseSwapchainImage.__doc__ = """
    Release a swapchain image.
"""
xrReleaseSwapchainImage.restype = Result
xrReleaseSwapchainImage.argtypes = [
    Swapchain,  # swapchain
    POINTER(SwapchainImageReleaseInfo),  # release_info
]

xrBeginSession = openxr_loader_library.xrBeginSession
xrBeginSession.__doc__ = """
    Begins an XrSession.
"""
xrBeginSession.restype = Result
xrBeginSession.argtypes = [
    Session,  # session
    POINTER(SessionBeginInfo),  # begin_info
]

xrEndSession = openxr_loader_library.xrEndSession
xrEndSession.__doc__ = """
    Ends an XrSession.
"""
xrEndSession.restype = Result
xrEndSession.argtypes = [
    Session,  # session
]

xrRequestExitSession = openxr_loader_library.xrRequestExitSession
xrRequestExitSession.__doc__ = """
    Request to exit a running session.
"""
xrRequestExitSession.restype = Result
xrRequestExitSession.argtypes = [
    Session,  # session
]

xrWaitFrame = openxr_loader_library.xrWaitFrame
xrWaitFrame.__doc__ = """
    Frame timing function.
"""
xrWaitFrame.restype = Result
xrWaitFrame.argtypes = [
    Session,  # session
    POINTER(FrameWaitInfo),  # frame_wait_info
    POINTER(FrameState),  # frame_state
]

xrBeginFrame = openxr_loader_library.xrBeginFrame
xrBeginFrame.__doc__ = """
    Marks a frame.
"""
xrBeginFrame.restype = Result
xrBeginFrame.argtypes = [
    Session,  # session
    POINTER(FrameBeginInfo),  # frame_begin_info
]

xrEndFrame = openxr_loader_library.xrEndFrame
xrEndFrame.__doc__ = """
    Marks a frame.
"""
xrEndFrame.restype = Result
xrEndFrame.argtypes = [
    Session,  # session
    POINTER(FrameEndInfo),  # frame_end_info
]

xrLocateViews = openxr_loader_library.xrLocateViews
xrLocateViews.__doc__ = """
    Gets view and projection info.
"""
xrLocateViews.restype = Result
xrLocateViews.argtypes = [
    Session,  # session
    POINTER(ViewLocateInfo),  # view_locate_info
    POINTER(ViewState),  # view_state
    c_uint32,  # view_capacity_input
    POINTER(c_uint32),  # view_count_output
    POINTER(View),  # views
]

xrStringToPath = openxr_loader_library.xrStringToPath
xrStringToPath.__doc__ = """
    Converts a string to a semantic path.
"""
xrStringToPath.restype = Result
xrStringToPath.argtypes = [
    Instance,  # instance
    c_char_p,  # path_string
    POINTER(Path),  # path
]

xrPathToString = openxr_loader_library.xrPathToString
xrPathToString.__doc__ = """
    Converts a semantic path to a string.
"""
xrPathToString.restype = Result
xrPathToString.argtypes = [
    Instance,  # instance
    Path,  # path
    c_uint32,  # buffer_capacity_input
    POINTER(c_uint32),  # buffer_count_output
    c_char_p,  # buffer
]

xrCreateActionSet = openxr_loader_library.xrCreateActionSet
xrCreateActionSet.__doc__ = """
    Creates an XrActionSet.
"""
xrCreateActionSet.restype = Result
xrCreateActionSet.argtypes = [
    Instance,  # instance
    POINTER(ActionSetCreateInfo),  # create_info
    POINTER(ActionSet),  # action_set
]

xrDestroyActionSet = openxr_loader_library.xrDestroyActionSet
xrDestroyActionSet.__doc__ = """
    Destroys an XrActionSet.
"""
xrDestroyActionSet.restype = Result
xrDestroyActionSet.argtypes = [
    ActionSet,  # action_set
]

xrCreateAction = openxr_loader_library.xrCreateAction
xrCreateAction.__doc__ = """
    Creates an XrAction.
"""
xrCreateAction.restype = Result
xrCreateAction.argtypes = [
    ActionSet,  # action_set
    POINTER(ActionCreateInfo),  # create_info
    POINTER(Action),  # action
]

xrDestroyAction = openxr_loader_library.xrDestroyAction
xrDestroyAction.__doc__ = """
    Destroys an XrAction.
"""
xrDestroyAction.restype = Result
xrDestroyAction.argtypes = [
    Action,  # action
]

xrSuggestInteractionProfileBindings = openxr_loader_library.xrSuggestInteractionProfileBindings
xrSuggestInteractionProfileBindings.__doc__ = """
    Sets the application-suggested bindings for the interaction
    profile.
"""
xrSuggestInteractionProfileBindings.restype = Result
xrSuggestInteractionProfileBindings.argtypes = [
    Instance,  # instance
    POINTER(InteractionProfileSuggestedBinding),  # suggested_bindings
]

xrAttachSessionActionSets = openxr_loader_library.xrAttachSessionActionSets
xrAttachSessionActionSets.__doc__ = """
    Attaches action sets to a given session.
"""
xrAttachSessionActionSets.restype = Result
xrAttachSessionActionSets.argtypes = [
    Session,  # session
    POINTER(SessionActionSetsAttachInfo),  # attach_info
]

xrGetCurrentInteractionProfile = openxr_loader_library.xrGetCurrentInteractionProfile
xrGetCurrentInteractionProfile.__doc__ = """
    Gets the current interaction profile for a top level user paths.
"""
xrGetCurrentInteractionProfile.restype = Result
xrGetCurrentInteractionProfile.argtypes = [
    Session,  # session
    Path,  # top_level_user_path
    POINTER(InteractionProfileState),  # interaction_profile
]

xrGetActionStateBoolean = openxr_loader_library.xrGetActionStateBoolean
xrGetActionStateBoolean.__doc__ = """
    Gets boolean action state.
"""
xrGetActionStateBoolean.restype = Result
xrGetActionStateBoolean.argtypes = [
    Session,  # session
    POINTER(ActionStateGetInfo),  # get_info
    POINTER(ActionStateBoolean),  # state
]

xrGetActionStateFloat = openxr_loader_library.xrGetActionStateFloat
xrGetActionStateFloat.__doc__ = """
    Gets a floating point action state.
"""
xrGetActionStateFloat.restype = Result
xrGetActionStateFloat.argtypes = [
    Session,  # session
    POINTER(ActionStateGetInfo),  # get_info
    POINTER(ActionStateFloat),  # state
]

xrGetActionStateVector2f = openxr_loader_library.xrGetActionStateVector2f
xrGetActionStateVector2f.__doc__ = """
    Gets 2D float vector action state.
"""
xrGetActionStateVector2f.restype = Result
xrGetActionStateVector2f.argtypes = [
    Session,  # session
    POINTER(ActionStateGetInfo),  # get_info
    POINTER(ActionStateVector2f),  # state
]

xrGetActionStatePose = openxr_loader_library.xrGetActionStatePose
xrGetActionStatePose.__doc__ = """
    Gets metadata from a pose action.
"""
xrGetActionStatePose.restype = Result
xrGetActionStatePose.argtypes = [
    Session,  # session
    POINTER(ActionStateGetInfo),  # get_info
    POINTER(ActionStatePose),  # state
]

xrSyncActions = openxr_loader_library.xrSyncActions
xrSyncActions.__doc__ = """
    Updates the current state of input actions.
"""
xrSyncActions.restype = Result
xrSyncActions.argtypes = [
    Session,  # session
    POINTER(ActionsSyncInfo),  # sync_info
]

xrEnumerateBoundSourcesForAction = openxr_loader_library.xrEnumerateBoundSourcesForAction
xrEnumerateBoundSourcesForAction.__doc__ = """
    Queries the bound input sources for an action.
"""
xrEnumerateBoundSourcesForAction.restype = Result
xrEnumerateBoundSourcesForAction.argtypes = [
    Session,  # session
    POINTER(BoundSourcesForActionEnumerateInfo),  # enumerate_info
    c_uint32,  # source_capacity_input
    POINTER(c_uint32),  # source_count_output
    POINTER(Path),  # sources
]

xrGetInputSourceLocalizedName = openxr_loader_library.xrGetInputSourceLocalizedName
xrGetInputSourceLocalizedName.__doc__ = """
    Gets a localized source name.
"""
xrGetInputSourceLocalizedName.restype = Result
xrGetInputSourceLocalizedName.argtypes = [
    Session,  # session
    POINTER(InputSourceLocalizedNameGetInfo),  # get_info
    c_uint32,  # buffer_capacity_input
    POINTER(c_uint32),  # buffer_count_output
    c_char_p,  # buffer
]

xrApplyHapticFeedback = openxr_loader_library.xrApplyHapticFeedback
xrApplyHapticFeedback.__doc__ = """
    Apply haptic feedback.
"""
xrApplyHapticFeedback.restype = Result
xrApplyHapticFeedback.argtypes = [
    Session,  # session
    POINTER(HapticActionInfo),  # haptic_action_info
    POINTER(HapticBaseHeader),  # haptic_feedback
]

xrStopHapticFeedback = openxr_loader_library.xrStopHapticFeedback
xrStopHapticFeedback.__doc__ = """
    Stop haptic feedback.
"""
xrStopHapticFeedback.restype = Result
xrStopHapticFeedback.argtypes = [
    Session,  # session
    POINTER(HapticActionInfo),  # haptic_action_info
]


__all__ = [
    "xrAcquireSwapchainImage",
    "xrApplyHapticFeedback",
    "xrAttachSessionActionSets",
    "xrBeginFrame",
    "xrBeginSession",
    "xrCreateAction",
    "xrCreateActionSet",
    "xrCreateActionSpace",
    "xrCreateInstance",
    "xrCreateReferenceSpace",
    "xrCreateSession",
    "xrCreateSwapchain",
    "xrDestroyAction",
    "xrDestroyActionSet",
    "xrDestroyInstance",
    "xrDestroySession",
    "xrDestroySpace",
    "xrDestroySwapchain",
    "xrEndFrame",
    "xrEndSession",
    "xrEnumerateApiLayerProperties",
    "xrEnumerateBoundSourcesForAction",
    "xrEnumerateEnvironmentBlendModes",
    "xrEnumerateInstanceExtensionProperties",
    "xrEnumerateReferenceSpaces",
    "xrEnumerateSwapchainFormats",
    "xrEnumerateSwapchainImages",
    "xrEnumerateViewConfigurationViews",
    "xrEnumerateViewConfigurations",
    "xrGetActionStateBoolean",
    "xrGetActionStateFloat",
    "xrGetActionStatePose",
    "xrGetActionStateVector2f",
    "xrGetCurrentInteractionProfile",
    "xrGetInputSourceLocalizedName",
    "xrGetInstanceProcAddr",
    "xrGetInstanceProperties",
    "xrGetReferenceSpaceBoundsRect",
    "xrGetSystem",
    "xrGetSystemProperties",
    "xrGetViewConfigurationProperties",
    "xrLocateSpace",
    "xrLocateViews",
    "xrPathToString",
    "xrPollEvent",
    "xrReleaseSwapchainImage",
    "xrRequestExitSession",
    "xrResultToString",
    "xrStopHapticFeedback",
    "xrStringToPath",
    "xrStructureTypeToString",
    "xrSuggestInteractionProfileBindings",
    "xrSyncActions",
    "xrWaitFrame",
    "xrWaitSwapchainImage",
]



import abc
import json
import os
import tempfile
from ctypes import c_char_p, POINTER

from ..enums import Result
from .layer_path import add_folder_to_api_layer_path, py_layer_library_path
from .loader_interfaces import PFN_xrNegotiateLoaderApiLayerInterface, NegotiateLoaderInfo, \
    NegotiateApiLayerRequest
from .raw_functions import insertXrApiLayer


class DynamicApiLayerBase(abc.ABC):
    """Base class for temporary dynamic runtime python OpenXR API layers."""

    def __init__(self, name: str, description: str = "", json_path=None) -> None:
        self._name = name
        self.description = description
        self.json_path = json_path
        # store pointer to keep it alive
        self.p_negotiate_fn = PFN_xrNegotiateLoaderApiLayerInterface(
            self._negotiate_loader_api_layer_interface)
        insertXrApiLayer(self.name.encode(), len(self.name), self.p_negotiate_fn)
        # Store json file in temporary folder
        if self.json_path is None:
            self.json_dir = tempfile.TemporaryDirectory()
            self.json_dir.cleanup()  # So it will be deleted later  TODO: those folders remain
            self.json_path = self.json_dir.name
        os.mkdir(self.json_path)
        add_folder_to_api_layer_path(self.json_path)
        json_file_name = f"{self.json_path}/{self.name}.json"
        with open(json_file_name, "w") as layer_json:
            layer_json.write(json.dumps(
                {
                    "file_format_version": "1.0.0",
                    "api_layer": {
                        "name": f"{self.name}",
                        "library_path": f"{py_layer_library_path()}",
                        "api_version": "1.0",
                        "implementation_version": "1",
                        "description": f"{self.description}"
                    }
                }, indent=4))
            layer_json.write("\n")

    @property
    def name(self) -> str:
        return self._name

    def _negotiate_loader_api_layer_interface(
            self,
            loader_info: NegotiateLoaderInfo,
            layer_name: c_char_p,
            api_layer_request: POINTER(NegotiateApiLayerRequest),
    ) -> int:
        """Lower-level function that will be passed to the C loader library."""
        try:
            # noinspection PyUnresolvedReferences
            result = self.negotiate_loader_api_layer_interface(
                loader_info.contents,
                layer_name.decode(),
                api_layer_request.contents,
            )
            return result.value
        except Exception:
            return Result.ERROR_INITIALIZATION_FAILED.value

    @abc.abstractmethod
    def negotiate_loader_api_layer_interface(
            self,
            loader_info: NegotiateLoaderInfo,
            layer_name: str,
            api_layer_request: NegotiateApiLayerRequest,
    ) -> Result:
        """
        Override this method in a derived class to create your own temporary dynamic OpenXR API layer.

        If this layer is able to support the request, it must: return xr.Result.SUCCESS and:
            Fill in pname:layerRequest→pname:layerInterfaceVersion with the API layer interface version it desires to support.
            Fill in pname:layerRequest→pname:layerApiVersion with the API version of OpenXR it will execute under.
            Fill in pname:layerRequest→pname:getInstanceProcAddr with a valid function pointer so that the loader can query function pointers to the remaining OpenXR commands supported by the API layer.
            Fill in pname:layerRequest→pname:createLayerInstance with a valid function pointer so that the loader can create the instance through the API layer call chain.

        Otherwise, it must: return XR_ERROR_INITIALIZATION_FAILED

        :param: loader_info: must be a valid pointer to a constant xr.NegotiateLoaderInfo structure.
        :param: layer_name: must be a string listing the name of an API layer which the loader is attempting to negotiate with.
        :param: api_layer_request: must be a valid pointer to a xr.NegotiateApiLayerRequest structure.
        :return: xr.Result.SUCCESS or xr.Result.ERROR_INITIALIZATION_FAILED
        """
        return Result.ERROR_INITIALIZATION_FAILED


__all__ = [
    "DynamicApiLayerBase",
]



from ctypes import byref, c_char_p, cast, memmove, POINTER, pointer, sizeof
import platform
import xr
from xr.raw_functions import xrGetInstanceProperties


class SteamVrLinuxDestroyInstanceLayer(xr.DynamicApiLayerBase):
    """
    Pure-python OpenXR API layer.
    Designed to work around SteamVR linux bug.
    https://steamcommunity.com/app/250820/discussions/8/3114772813482874960/
    https://github.com/cmbruns/pyopenxr/pull/66
    """
    def __init__(self):
        super().__init__(name="XR_APILAYER_PYOPENXR_Example_Layer")
        # Store function pointers to keep them alive
        self.p_xrGetInstanceProcAddr = xr.PFN_xrGetInstanceProcAddr(self.get_instance_proc_addr)
        self.p_xrCreateApiLayerInstance = xr.PFN_xrCreateApiLayerInstance(self.create_api_layer_instance)
        self.nextXrGetInstanceProcAddr = None
        self.nextXrDestroyInstance = xr.PFN_xrDestroyInstance()
        self.p_xrDestroyInstance = xr.PFN_xrDestroyInstance(self.destroy_instance)
        self.is_steamvr = False

    def create_api_layer_instance(
        self,
        info: POINTER(xr.InstanceCreateInfo),
        api_layer_info: POINTER(xr.ApiLayerCreateInfo),
        instance: xr.Instance,
    ) -> xr.Result:
        # Dereference the next_info field for easier calling.
        next_info = api_layer_info.contents.next_info.contents
        self.nextXrGetInstanceProcAddr = next_info.next_get_instance_proc_addr
        # Let the instance be created.
        result = next_info.next_create_api_layer_instance(info, api_layer_info, instance)
        if xr.failed(result):
            return result
        # Remember the subsequent layer's xrDestroyInstance() function
        fn = xr.PFN_xrVoidFunction()
        result = self.nextXrGetInstanceProcAddr(
            instance,
            b"xrDestroyInstance",  # Use bytes object not string because this is low level C API.
            pointer(fn),
        )
        self.nextXrDestroyInstance = cast(fn, xr.PFN_xrDestroyInstance)
        if xr.failed(result):
            return result
        #
        return result

    def get_instance_proc_addr(
        self,
        instance: xr.Instance,
        name: c_char_p,
        function: POINTER(xr.PFN_xrVoidFunction),
    ) -> xr.Result:
        # Intercept calls to xrDestroyInstance
        if name == b"xrDestroyInstance":
            # TODO: Is there a neater way than memmove?
            memmove(function, byref(self.p_xrDestroyInstance), sizeof(xr.PFN_xrVoidFunction))
            return xr.Result.SUCCESS
        # Delegate all other function calls to the subsequent API layer
        return self.nextXrGetInstanceProcAddr(instance, name, function)

    # Signature must exactly match the low-level signature of xrDestroyInstance
    def destroy_instance(self, instance: xr.Instance) -> xr.Result:
        instance_properties = xr.InstanceProperties()
        xrGetInstanceProperties(instance, byref(instance_properties))
        is_steam_vr = instance_properties.runtime_name == b"SteamVR/OpenXR"
        if is_steam_vr and platform.system() == "Linux":
            return xr.Result.SUCCESS  # Elide call to xrDestroyInstance
        # Propagate function call down into to the next API layer
        return self.nextXrDestroyInstance(instance)

    def negotiate_loader_api_layer_interface(
            self,
            loader_info: xr.NegotiateLoaderInfo,
            layer_name: str,
            api_layer_request: xr.NegotiateApiLayerRequest,
    ) -> xr.Result:
        """
        Set up our layer to intercept OpenXR function calls.
        :param loader_info:
        :param layer_name: The name of an API layer which the loader is attempting to negotiate with.
        :param api_layer_request: Fill in this information.
        :return: xr.Result.SUCCESS or xr.Result.ERROR_INITIALIZATION_FAILED
        """
        # TODO: compare versions in loader_info
        api_layer_request.layer_interface_version = xr.CURRENT_LOADER_API_LAYER_VERSION
        api_layer_request.layer_api_version = xr.XR_CURRENT_API_VERSION
        api_layer_request.get_instance_proc_addr = self.p_xrGetInstanceProcAddr
        api_layer_request.create_api_layer_instance = self.p_xrCreateApiLayerInstance
        return xr.Result.SUCCESS


if __name__ == "__main__":
    dle = SteamVrLinuxDestroyInstanceLayer()
    assert dle.name in xr.enumerate_api_layer_properties()
    instance_handle = xr.create_instance(xr.InstanceCreateInfo(
        enabled_api_layer_names=[
            # xr.XR_APILAYER_LUNARG_api_dump_NAME,
            # xr.XR_APILAYER_LUNARG_core_validation_NAME,
            dle.name,
            # TODO: adding layers after causes problems
            # xr.XR_APILAYER_LUNARG_core_validation_NAME,
            # xr.XR_APILAYER_LUNARG_api_dump_NAME,
        ],
    ))
    xr.destroy_instance(instance_handle)  # Should be no-op on Linux/SteamVR with layer active


__all__ = [
    "SteamVrLinuxDestroyInstanceLayer",
]



import os
import pkg_resources
import platform


def add_folder_to_api_layer_path(folder_name: str):
    starting_api_path = os.getenv("XR_API_LAYER_PATH")
    if starting_api_path is None or len(starting_api_path) < 1:
        os.environ["XR_API_LAYER_PATH"] = folder_name
    elif folder_name in starting_api_path.split(os.pathsep):
        pass  # It's already there
    else:
        # pro-tip: os.pathsep is very different from os.path.sep
        os.environ["XR_API_LAYER_PATH"] += f"{os.pathsep}{folder_name}"


def expose_packaged_api_layers():
    """
    Make pre-packaged layers available to the openxr loader
    """
    if platform.system() == "Windows":
        local_path = os.path.abspath(pkg_resources.resource_filename(__name__, "windows"))
    elif platform.system() == "Linux":
        local_path = os.path.abspath(pkg_resources.resource_filename(__name__, "linux"))
    else:
        raise NotImplementedError
    add_folder_to_api_layer_path(local_path)


def py_layer_library_path() -> str:
    """Path to a shared library file used for dynamic API layer dispatch."""
    if platform.system() == "Windows":
        package = "xr.api_layer.windows"
        name = "XrApiLayer_python.dll"
    elif platform.system() == "Linux":
        package = "xr.api_layer.linux"
        name = "libXrApiLayer_python.so"
    else:
        raise NotImplementedError
    path = pkg_resources.resource_filename(package, name)
    return path


__all__ = [
    "add_folder_to_api_layer_path",
    "expose_packaged_api_layers",
    "py_layer_library_path",
]



"""
file array_field.py

Helper functions for Structures containing arrays specified as length/pointer pairs.
"""

import enum
from ctypes import Array, c_char_p, cast, POINTER, pointer, Structure
from typing import TypeVar, Union, Sequence, Optional


E = TypeVar("E")
ArrayFieldParamType = Union[
    None,
    POINTER,
    E,
    Array,
    Sequence[E],
]

BaseArrayFieldParamType = Union[
    None,
    POINTER,
    Array,
    Sequence,
]

StringArrayFieldParamType = Union[
    None,
    POINTER(c_char_p),
    c_char_p,
    Array,
    Sequence[str],
]


class ArrayFlavor(enum.Enum):
    ARRAY = 1,
    STRING = 2,
    BASE_HEADER = 3,


def array_field_helper(
        element_type: type,
        count: Optional[int] = None,
        array: ArrayFieldParamType["element_type"] = None,
        flavor: ArrayFlavor = ArrayFlavor.ARRAY
) -> (int, POINTER):
    """Helper function for pythonic interface to sequence fields"""
    # Default construction yields a null pointer and a zero length
    if array is None:
        count = 0
    # Construction from pointer usually requires an explicit length.
    # ...but default to 1 if length is not provided.
    elif isinstance(array, POINTER(element_type)):
        if count is None:
            count = 1
    elif isinstance(array, element_type):
        count = 1
        array = pointer(array)
    elif isinstance(array, Array):
        count = len(array)
    else:
        # Copy array into a ctypes array
        count = len(array)
        if flavor == ArrayFlavor.ARRAY:
            array = (element_type * count)(*array)
        elif flavor == ArrayFlavor.STRING:
            array = (element_type * count)(*[s.encode() for s in array])
        elif flavor == ArrayFlavor.BASE_HEADER:
            array = (element_type * count)(*[cast(p, element_type) for p in array])
    return count, array


def base_array_field_helper(
        element_type: type,
        count: Optional[int] = None,
        array: BaseArrayFieldParamType = None,
) -> (int, POINTER):
    """Helper function for pythonic interface to sequence fields"""
    return array_field_helper(element_type, count, array, flavor=ArrayFlavor.BASE_HEADER)


def string_array_field_helper(
        count: Optional[int] = None,
        array: StringArrayFieldParamType = None,
) -> (int, POINTER(c_char_p)):
    """Helper function for pythonic interface to sequence fields"""
    return array_field_helper(c_char_p, count, array, flavor=ArrayFlavor.STRING)


__all__ = [
    "array_field_helper",
    "ArrayFieldParamType",
    "base_array_field_helper",
    "BaseArrayFieldParamType",
    "string_array_field_helper",
    "StringArrayFieldParamType",
]



import time
from ctypes import byref, c_int32, c_void_p, cast, POINTER, pointer, Structure

import xr
from .enums import *
from .exception import *
from .typedefs import *
from .functions import *
from .opengl_graphics import OpenGLGraphics


class SwapchainStruct(Structure):
    _fields_ = [
        ("handle", Swapchain),
        ("width", c_int32),
        ("height", c_int32),
    ]


class ContextObject(object):
    def __init__(
            self,
            instance_create_info: InstanceCreateInfo = InstanceCreateInfo(),
            session_create_info: SessionCreateInfo = SessionCreateInfo(),
            reference_space_create_info: ReferenceSpaceCreateInfo = ReferenceSpaceCreateInfo(),
            view_configuration_type: ViewConfigurationType = ViewConfigurationType.PRIMARY_STEREO,
            environment_blend_mode=EnvironmentBlendMode.OPAQUE,
            form_factor=FormFactor.HEAD_MOUNTED_DISPLAY,
    ):
        self._instance_create_info = instance_create_info
        self.instance = None
        self._session_create_info = session_create_info
        self.session = None
        self.session_state = SessionState.IDLE
        self._reference_space_create_info = reference_space_create_info
        self.view_configuration_type = view_configuration_type
        self.environment_blend_mode = environment_blend_mode
        self.form_factor = form_factor
        self.graphics = None
        self.graphics_binding_pointer = None
        self.action_sets = []
        self.render_layers = []
        self.swapchains = []
        self.swapchain_image_ptr_buffers = []
        self.swapchain_image_buffers = []  # Keep alive
        self.exit_render_loop = False
        self.request_restart = False  # TODO: do like hello_xr
        self.session_is_running = False

    def __enter__(self):
        self.instance = create_instance(
            create_info=self._instance_create_info,
        )
        self.system_id = get_system(
            instance=self.instance,
            get_info=SystemGetInfo(
                form_factor=self.form_factor,
            ),
        )

        if self._session_create_info.next is None:
            self.graphics = OpenGLGraphics(
                instance=self.instance,
                system=self.system_id,
                title=self._instance_create_info.application_info.application_name.decode()
            )
            self.graphics_binding_pointer = cast(pointer(self.graphics.graphics_binding), c_void_p)
            self._session_create_info.next = self.graphics_binding_pointer
        else:
            self.graphics_binding_pointer = self._session_create_info.next

        self._session_create_info.system_id = self.system_id
        self.session = create_session(
            instance=self.instance,
            create_info=self._session_create_info,
        )
        self.space = create_reference_space(
            session=self.session,
            create_info=self._reference_space_create_info
        )
        self.default_action_set = create_action_set(
            instance=self.instance,
            create_info=ActionSetCreateInfo(
                action_set_name="default_action_set",
                localized_action_set_name="Default Action Set",
                priority=0,
            ),
        )
        self.action_sets.append(self.default_action_set)

        # Create swapchains
        config_views = xr.enumerate_view_configuration_views(
            instance=self.instance,
            system_id=self.system_id,
            view_configuration_type=self.view_configuration_type,
        )
        self.graphics.initialize_resources()
        swapchain_formats = xr.enumerate_swapchain_formats(self.session)
        color_swapchain_format = self.graphics.select_color_swapchain_format(swapchain_formats)
        # Create a swapchain for each view.
        self.swapchains.clear()
        self.swapchain_image_buffers.clear()
        self.swapchain_image_ptr_buffers.clear()
        for vp in config_views:
            # Create the swapchain.
            swapchain_create_info = xr.SwapchainCreateInfo(
                array_size=1,
                format=color_swapchain_format,
                width=vp.recommended_image_rect_width,
                height=vp.recommended_image_rect_height,
                mip_count=1,
                face_count=1,
                sample_count=vp.recommended_swapchain_sample_count,
                usage_flags=xr.SwapchainUsageFlags.SAMPLED_BIT | xr.SwapchainUsageFlags.COLOR_ATTACHMENT_BIT,
            )
            swapchain = SwapchainStruct(
                xr.create_swapchain(
                    session=self.session,
                    create_info=swapchain_create_info,
                ),
                swapchain_create_info.width,
                swapchain_create_info.height,
            )
            self.swapchains.append(swapchain)
            swapchain_image_buffer = xr.enumerate_swapchain_images(
                swapchain=swapchain.handle,
                element_type=self.graphics.swapchain_image_type,
            )
            # Keep the buffer alive by moving it into the list of buffers.
            self.swapchain_image_buffers.append(swapchain_image_buffer)
            capacity = len(swapchain_image_buffer)
            swapchain_image_ptr_buffer = (POINTER(xr.SwapchainImageBaseHeader) * capacity)()
            for ix in range(capacity):
                swapchain_image_ptr_buffer[ix] = cast(
                    byref(swapchain_image_buffer[ix]),
                    POINTER(xr.SwapchainImageBaseHeader))
            self.swapchain_image_ptr_buffers.append(swapchain_image_ptr_buffer)
        self.graphics.make_current()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.default_action_set is not None:
            destroy_action_set(self.default_action_set)
            self.default_action_set = None
        if self.space is not None:
            destroy_space(self.space)
            self.space = None
        if self.session is not None:
            destroy_session(self.session)
            self.session = None
        if self.graphics is not None:
            self.graphics.destroy()
            self.graphics = None
        if self.instance is not None:
            destroy_instance(self.instance)
            self.instance = None

    def frame_loop(self):
        attach_session_action_sets(
            session=self.session,
            attach_info=SessionActionSetsAttachInfo(
                count_action_sets=len(self.action_sets),
                action_sets=(ActionSet * len(self.action_sets))(
                    *self.action_sets
                )
            ),
        )
        while True:
            window_closed = self.graphics.poll_events()
            if window_closed:
                self.exit_render_loop = True
                break
            self.exit_render_loop = False
            self.poll_xr_events()
            if self.exit_render_loop:
                break
            if self.session_is_running:
                if self.session_state in (
                        SessionState.READY,
                        SessionState.SYNCHRONIZED,
                        SessionState.VISIBLE,
                        SessionState.FOCUSED,
                ):
                    frame_state = wait_frame(self.session)
                    begin_frame(self.session)
                    self.render_layers = []
                    self.graphics.make_current()

                    yield frame_state

                    end_frame(
                        self.session,
                        frame_end_info=FrameEndInfo(
                            display_time=frame_state.predicted_display_time,
                            environment_blend_mode=self.environment_blend_mode,
                            layers=self.render_layers,
                        )
                    )
            else:
                # Throttle loop since xrWaitFrame won't be called.
                time.sleep(0.250)

    def poll_xr_events(self):
        self.exit_render_loop = False
        self.request_restart = False
        while True:
            try:
                event_buffer = poll_event(self.instance)
                event_type = StructureType(event_buffer.type)
                if event_type == StructureType.EVENT_DATA_INSTANCE_LOSS_PENDING:
                    # still handle rest of the events instead of immediately quitting
                    self.exit_render_loop = True
                    self.request_restart = True
                elif event_type == StructureType.EVENT_DATA_SESSION_STATE_CHANGED \
                        and self.session is not None:
                    event = cast(
                        byref(event_buffer),
                        POINTER(EventDataSessionStateChanged)).contents
                    self.session_state = SessionState(event.state)
                    if self.session_state == SessionState.READY:
                        begin_session(
                            session=self.session,
                            begin_info=SessionBeginInfo(
                                self.view_configuration_type,
                            ),
                        )
                        self.session_is_running = True
                    elif self.session_state == SessionState.STOPPING:
                        self.session_is_running = False
                        xr.end_session(self.session)
                    elif self.session_state == SessionState.EXITING:
                        self.exit_render_loop = True
                        self.request_restart = False
                    elif self.session_state == SessionState.LOSS_PENDING:
                        self.exit_render_loop = True
                        self.request_restart = True
                elif event_type == StructureType.EVENT_DATA_VIVE_TRACKER_CONNECTED_HTCX:
                    vive_tracker_connected = cast(byref(event_buffer), POINTER(EventDataViveTrackerConnectedHTCX)).contents
                    paths = vive_tracker_connected.paths.contents
                    persistent_path_str = xr.path_to_string(self.instance, paths.persistent_path)
                    # print(f"Vive Tracker connected: {persistent_path_str}")
                    if paths.role_path != xr.NULL_PATH:
                        role_path_str = xr.path_to_string(self.instance, paths.role_path)
                        # print(f" New role is: {role_path_str}")
                    else:
                        # print(f" No role path.")
                        pass
                elif event_type == StructureType.EVENT_DATA_INTERACTION_PROFILE_CHANGED:
                    # print("data interaction profile changed")
                    # TODO:
                    pass
            except EventUnavailable:
                break

    def view_loop(self, frame_state):
        if frame_state.should_render:
            layer = xr.CompositionLayerProjection(space=self.space)
            projection_layer_views = (xr.CompositionLayerProjectionView * 2)(
                xr.CompositionLayerProjectionView(),
                xr.CompositionLayerProjectionView())
            view_state, views = xr.locate_views(
                session=self.session,
                view_locate_info=xr.ViewLocateInfo(
                    view_configuration_type=self.view_configuration_type,
                    display_time=frame_state.predicted_display_time,
                    space=self.space,
                )
            )
            vsf = view_state.view_state_flags
            if (vsf & xr.VIEW_STATE_POSITION_VALID_BIT == 0
                    or vsf & xr.VIEW_STATE_ORIENTATION_VALID_BIT == 0):
                return  # There are no valid tracking poses for the views.
            for view_index, view in enumerate(views):
                view_swapchain = self.swapchains[view_index]
                swapchain_image_index = xr.acquire_swapchain_image(
                    swapchain=view_swapchain.handle,
                    acquire_info=xr.SwapchainImageAcquireInfo(),
                )
                xr.wait_swapchain_image(
                    swapchain=view_swapchain.handle,
                    wait_info=xr.SwapchainImageWaitInfo(timeout=xr.INFINITE_DURATION),
                )
                layer_view = projection_layer_views[view_index]
                assert layer_view.type == xr.StructureType.COMPOSITION_LAYER_PROJECTION_VIEW
                layer_view.pose = view.pose
                layer_view.fov = view.fov
                layer_view.sub_image.swapchain = view_swapchain.handle
                layer_view.sub_image.image_rect.offset[:] = [0, 0]
                layer_view.sub_image.image_rect.extent[:] = [
                    view_swapchain.width, view_swapchain.height, ]
                swapchain_image_ptr = self.swapchain_image_ptr_buffers[view_index][swapchain_image_index]
                swapchain_image = cast(swapchain_image_ptr, POINTER(xr.SwapchainImageOpenGLKHR)).contents
                assert layer_view.sub_image.image_array_index == 0  # texture arrays not supported.
                color_texture = swapchain_image.image
                self.graphics.begin_frame(layer_view, color_texture)

                yield view

                self.graphics.end_frame()
                xr.release_swapchain_image(
                    swapchain=view_swapchain.handle,
                    release_info=xr.SwapchainImageReleaseInfo()
                )
            layer.views = projection_layer_views
            self.render_layers.append(byref(layer))


__all__ = [
    "ContextObject",
]



import ctypes
from ctypes import c_char, c_char_p, c_uint32, c_void_p, CFUNCTYPE, POINTER, c_size_t

from ..enums import EnumBase, Result
from ..typedefs import (
    InstanceCreateInfo,
    Instance,
    PFN_xrGetInstanceProcAddr,
    VersionNumber
)
from ..version import Version

# TODO: generate the header information from C automatically

# Loader/API Layer Interface versions
#  1 - First version, introduces negotiation structure and functions
CURRENT_LOADER_API_LAYER_VERSION = 1

MAX_API_LAYER_NAME_SIZE = 256
API_LAYER_MAX_SETTINGS_PATH_SIZE = 512


# Version negotiation values
class LoaderInterfaceStructs(EnumBase):
    UNINITIALIZED = 0,  # corrected spelling error in the C API.
    LOADER_INFO = 1,
    API_LAYER_REQUEST = 2,
    RUNTIME_REQUEST = 3,
    API_LAYER_CREATE_INFO = 4,
    API_LAYER_NEXT_INFO = 5,

    def __str__(self):
        return f"xr.{super().__str__()}"


class NegotiateLoaderInfo(ctypes.Structure):
    _fields_ = [
        ("struct_type", LoaderInterfaceStructs.ctype()),  # XR_LOADER_INTERFACE_STRUCT_LOADER_INFO
        ("struct_version", c_uint32),  # XR_LOADER_INFO_STRUCT_VERSION
        ("struct_size", c_size_t),  # sizeof(XrNegotiateLoaderInfo)
        ("min_interface_version", c_uint32),
        ("max_interface_version", c_uint32),
        ("min_api_version", VersionNumber),
        ("max_api_version", VersionNumber),
    ]

    def __str__(self) -> str:
        return ("NegotiateLoaderInfo("
                # Skipping the type enum, which is redundant
                f"struct_version={Version(self.struct_version)}"
                f", struct_size={self.struct_size}"
                f", min_interface_version={self.min_interface_version}"
                f", max_interface_version={self.max_interface_version}"
                f", min_api_version={Version(self.min_api_version)}"
                f", max_api_version={Version(self.max_api_version)}"
                ")")


# "forward" declaration
class ApiLayerNextInfo(ctypes.Structure):
    pass  # Fields are defined further below, for self-reference and declaration order reasons


class ApiLayerCreateInfo(ctypes.Structure):
    _fields_ = [
        ("struct_type", LoaderInterfaceStructs.ctype()),  # XR_LOADER_INTERFACE_STRUCT_API_LAYER_CREATE_INFO
        ("struct_version", c_uint32),  # XR_API_LAYER_CREATE_INFO_STRUCT_VERSION
        ("struct_size", c_size_t),  # sizeof(XrApiLayerCreateInfo)
        ("loader_instance", c_void_p),  # Pointer to the LoaderInstance class
        ("settings_file_location", c_char * API_LAYER_MAX_SETTINGS_PATH_SIZE),  # Location to the found settings file (or empty '\0')
        ("next_info", POINTER(ApiLayerNextInfo)),
    ]


PFN_xrCreateApiLayerInstance = CFUNCTYPE(
    Result.ctype(),  # return value
    POINTER(InstanceCreateInfo),  # info
    POINTER(ApiLayerCreateInfo),  # apiLayerInfo
    Instance,  # instance
)


# ApiLayerNextInfo class was "forward" declared earlier
ApiLayerNextInfo._fields_ = [
    ("struct_type", LoaderInterfaceStructs.ctype()),                # XR_LOADER_INTERFACE_STRUCT_API_LAYER_NEXT_INFO
    ("struct_version", c_uint32),                                   # XR_API_LAYER_NEXT_INFO_STRUCT_VERSION
    ("struct_size", c_size_t),                                      # sizeof(XrApiLayerNextInfo)
    ("layer_name", c_char * MAX_API_LAYER_NAME_SIZE),               # Name of API layer which should receive this info
    ("next_get_instance_proc_addr", PFN_xrGetInstanceProcAddr),        # Pointer to next API layer's xrGetInstanceProcAddr
    ("next_create_api_layer_instance", PFN_xrCreateApiLayerInstance),  # Pointer to next API layer's xrCreateApiLayerInstance
    ("next", POINTER(ApiLayerNextInfo)),                           # Pointer to the next API layer info in the sequence
]


class NegotiateApiLayerRequest(ctypes.Structure):
    _fields_ = [
        ("struct_type", LoaderInterfaceStructs.ctype()),   # XR_LOADER_INTERFACE_STRUCT_API_LAYER_REQUEST
        ("struct_version", c_uint32),                # XR_API_LAYER_INFO_STRUCT_VERSION
        ("struct_size", c_size_t),                   # sizeof(XrNegotiateApiLayerRequest)
        ("layer_interface_version", c_uint32),        # CURRENT_LOADER_API_LAYER_VERSION
        ("layer_api_version", VersionNumber),
        ("get_instance_proc_addr", PFN_xrGetInstanceProcAddr),
        ("create_api_layer_instance", PFN_xrCreateApiLayerInstance),
    ]

    def __str__(self) -> str:
        return ("NegotiateApiLayerRequest("
                # Skipping the type enum, which is redundant
                f"struct_version={Version(self.struct_version)}"
                f", struct_size={self.struct_size}"
                f", layer_interface_version={Version(self.layer_interface_version)}"
                f", layer_api_version={Version(self.layer_api_version)}"
                # Skip the function pointers when printing...
                ")")


PFN_xrNegotiateLoaderApiLayerInterface = CFUNCTYPE(
    Result.ctype(),  # return value
    POINTER(NegotiateLoaderInfo),  # loader_info
    c_char_p,  # api_layer_name
    POINTER(NegotiateApiLayerRequest),  # api_layer_request
)

__all__ = [
    "ApiLayerCreateInfo",
    "CURRENT_LOADER_API_LAYER_VERSION",
    "NegotiateLoaderInfo",
    "NegotiateApiLayerRequest",
    "PFN_xrCreateApiLayerInstance",
    "PFN_xrNegotiateLoaderApiLayerInterface",
]



import ctypes
import platform
from typing import Dict

import glfw
if platform.system() == "Windows":
    from OpenGL import WGL
    from .platform.windows import *
elif platform.system() == "Linux":
    from OpenGL import GLX
    from .platform.linux import *
from OpenGL import GL

from .enums import *
from .exception import *
from .typedefs import *
from .functions import *


class OpenGLGraphics(object):
    def __init__(
            self,
            instance: Instance,
            system: SystemId,
            title: str = "glfw OpenGL window",
    ) -> None:
        if not glfw.init():
            raise XrException("GLFW initialization failed")
        self.window_size = (64, 64)
        glfw.window_hint(glfw.VISIBLE, False)
        self.pxrGetOpenGLGraphicsRequirementsKHR = ctypes.cast(
            get_instance_proc_addr(
                instance=instance,
                name="xrGetOpenGLGraphicsRequirementsKHR",
            ),
            PFN_xrGetOpenGLGraphicsRequirementsKHR
        )
        self.graphics_requirements = GraphicsRequirementsOpenGLKHR()
        result = self.pxrGetOpenGLGraphicsRequirementsKHR(
            instance,
            system,
            ctypes.byref(self.graphics_requirements))
        result = check_result(Result(result))
        if result.is_exception():
            raise result
        glfw.window_hint(glfw.DOUBLEBUFFER, False)
        glfw.window_hint(glfw.CONTEXT_VERSION_MAJOR, 4)
        glfw.window_hint(glfw.CONTEXT_VERSION_MINOR, 5)
        glfw.window_hint(glfw.OPENGL_PROFILE, glfw.OPENGL_CORE_PROFILE)
        self.window = glfw.create_window(*self.window_size, title, None, None)
        if self.window is None:
            raise XrException("Failed to create GLFW window")
        glfw.make_context_current(self.window)
        # Attempt to disable vsync on the desktop window, or
        # it will interfere with the OpenXR frame loop timing
        glfw.swap_interval(0)
        self.graphics_binding = None
        if platform.system() == "Windows":
            self.graphics_binding = GraphicsBindingOpenGLWin32KHR()
            self.graphics_binding.h_dc = WGL.wglGetCurrentDC()
            self.graphics_binding.h_glrc = WGL.wglGetCurrentContext()
        elif platform.system() == "Linux":
            drawable = GLX.glXGetCurrentDrawable()
            context = GLX.glXGetCurrentContext()
            display = GLX.glXGetCurrentDisplay()
            self.graphics_binding = GraphicsBindingOpenGLXlibKHR(
                x_display=display,
                glx_drawable=drawable,
                glx_context=context,
            )
        else:
            raise NotImplementedError
        self.swapchain_framebuffer = None
        self.color_to_depth_map: Dict[int, int] = {}

    def __enter__(self):
        return self

    def __exit__(self, exception_type, value, traceback):
        self.destroy()

    def begin_frame(self, layer_view, color_texture):
        self.make_current()
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.swapchain_framebuffer)
        GL.glViewport(layer_view.sub_image.image_rect.offset.x,
                      layer_view.sub_image.image_rect.offset.y,
                      layer_view.sub_image.image_rect.extent.width,
                      layer_view.sub_image.image_rect.extent.height)
        depth_texture = self.get_depth_texture(color_texture)
        GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_COLOR_ATTACHMENT0, GL.GL_TEXTURE_2D, color_texture, 0)
        GL.glFramebufferTexture2D(GL.GL_FRAMEBUFFER, GL.GL_DEPTH_ATTACHMENT, GL.GL_TEXTURE_2D, depth_texture, 0)

    def destroy(self):
        self.make_current()
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)
        if self.swapchain_framebuffer is not None:
            GL.glDeleteFramebuffers(1, [self.swapchain_framebuffer, ])
            self.swapchain_framebuffer = None
        glfw.destroy_window(self.window)
        self.window = None
        glfw.terminate()

    @staticmethod
    def end_frame():
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)

    def get_depth_texture(self, color_texture) -> int:
        # If a depth-stencil view has already been created for this back-buffer, use it.
        if color_texture in self.color_to_depth_map:
            return self.color_to_depth_map[color_texture]
        # This back-buffer has no corresponding depth-stencil texture, so create one with matching dimensions.
        GL.glBindTexture(GL.GL_TEXTURE_2D, color_texture)
        width = GL.glGetTexLevelParameteriv(GL.GL_TEXTURE_2D, 0, GL.GL_TEXTURE_WIDTH)
        height = GL.glGetTexLevelParameteriv(GL.GL_TEXTURE_2D, 0, GL.GL_TEXTURE_HEIGHT)

        depth_texture = GL.glGenTextures(1)
        GL.glBindTexture(GL.GL_TEXTURE_2D, depth_texture)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, GL.GL_NEAREST)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, GL.GL_CLAMP_TO_EDGE)
        GL.glTexParameteri(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, GL.GL_CLAMP_TO_EDGE)
        GL.glTexImage2D(GL.GL_TEXTURE_2D, 0, GL.GL_DEPTH_COMPONENT32, width, height, 0, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT, None)
        self.color_to_depth_map[color_texture] = depth_texture
        return depth_texture

    def initialize_resources(self):
        self.make_current()
        self.swapchain_framebuffer = GL.glGenFramebuffers(1)

    def make_current(self):
        glfw.make_context_current(self.window)

    def poll_events(self) -> bool:
        glfw.poll_events()
        return glfw.window_should_close(self.window)

    @staticmethod
    def select_color_swapchain_format(runtime_formats):
        # List of supported color swapchain formats.
        supported_color_swapchain_formats = [
            GL.GL_RGB10_A2,
            GL.GL_RGBA16F,
            # The two below should only be used as a fallback, as they are linear color formats without enough bits for color
            # depth, thus leading to banding.
            GL.GL_RGBA8,
            GL.GL_RGBA8_SNORM,
            #
            GL.GL_SRGB8,  # Linux SteamVR beta 1.24.2 has only these...
            GL.GL_SRGB8_ALPHA8,
        ]
        for rf in runtime_formats:
            for sf in supported_color_swapchain_formats:
                if rf == sf:
                    return sf
        raise RuntimeError("No runtime swapchain format supported for color swapchain")

    @property
    def swapchain_image_type(self):
        return SwapchainImageOpenGLKHR


__all__ = [
    "OpenGLGraphics",
]
